# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import List, Union

from marshmallow import Schema

from ..._schema import PathAwareSchema
from .._schema.component import NodeType
from ..entities.node import InternalBaseNode


class Scope(InternalBaseNode):
    """Node of scope components in pipeline with specific run settings."""

    def __init__(self, **kwargs):
        kwargs.pop("type", None)
        super(Scope, self).__init__(type=NodeType.SCOPE, **kwargs)
        self._init = True
        self._adla_account_name = kwargs.pop("adla_account_name", None)
        self._scope_param = kwargs.pop("scope_param", None)
        self._custom_job_name_suffix = kwargs.pop("custom_job_name_suffix", None)
        self._priority = kwargs.pop("priority", None)
        self._auto_token = kwargs.pop("auto_token", None)
        self._tokens = kwargs.pop("tokens", None)
        self._vcp = kwargs.pop("vcp", None)
        self._init = False

    @property
    def adla_account_name(self) -> str:
        """The ADLA account name to use for the scope job.

        :return: ADLA account name
        :rtype: str
        """
        return self._adla_account_name

    @adla_account_name.setter
    def adla_account_name(self, value: str):
        self._adla_account_name = value

    @property
    def scope_param(self) -> str:
        """nebula command used when submit the scope job.

        :return: The nebula command
        :rtype: str
        """
        return self._scope_param

    @scope_param.setter
    def scope_param(self, value: str):
        self._scope_param = value

    @property
    def custom_job_name_suffix(self) -> str:
        """Optional string to append to scope job name.

        :return: The custom suffix
        :rtype: str
        """
        return self._custom_job_name_suffix

    @custom_job_name_suffix.setter
    def custom_job_name_suffix(self, value: str):
        self._custom_job_name_suffix = value

    @property
    def priority(self) -> int:
        """scope job priority.

        If set priority in scope_param, will override this setting.

        :return: The job priority
        :rtype: int
        """
        return self._priority

    @priority.setter
    def priority(self, value: int):
        self._priority = value

    @property
    def auto_token(self) -> int:
        """A predictor for estimating the peak resource usage of scope job.

        :return: auto token
        :rtype: int
        """
        return self._auto_token

    @auto_token.setter
    def auto_token(self, value: int):
        self._auto_token = value

    @property
    def tokens(self) -> int:
        """Standard token allocation.

        :return: The token allocation
        :rtype: int
        """
        return self._tokens

    @tokens.setter
    def tokens(self, value: int):
        self._tokens = value

    @property
    def vcp(self) -> float:
        """Standard VC percent allocation; should be a float between 0 and 1.

        :return: The VC allocation
        :rtype: float
        """
        return self._vcp

    @vcp.setter
    def vcp(self, value: float):
        self._vcp = value

    @classmethod
    def _picked_fields_from_dict_to_rest_object(cls) -> List[str]:
        return ["custom_job_name_suffix", "scope_param", "adla_account_name", "priority", "auto_token", "tokens", "vcp"]

    @classmethod
    def _create_schema_for_validation(cls, context) -> Union[PathAwareSchema, Schema]:
        from .._schema.node import ScopeSchema

        return ScopeSchema(context=context)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Any, List
from azure.ai.ml.entities._mixins import RestTranslatableMixin
from azure.ai.ml._restclient.v2024_10_01_preview.models import JobResources as RestJobResources


class JobResources(RestTranslatableMixin):
    """Resource configuration for a job.

    This class should not be instantiated directly. Instead, use its subclasses.
    """

    def __init__(self, *, instance_types: List[str]) -> None:
        self.instance_types = instance_types

    def _to_rest_object(self) -> Any:
        return RestJobResources(instance_types=self.instance_types)

    @classmethod
    def _from_rest_object(cls, obj: RestJobResources) -> "JobResources":
        job_resources = cls(instance_types=obj.instance_types)
        return job_resources

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, JobResources):
            return NotImplemented
        return self.instance_types == other.instance_types

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from azure.ai.ml._utils._arm_id_utils import AzureResourceId


class MaterializationStore:
    """Materialization Store

    :param type: The type of the materialization store.
    :type type: str
    :param target: The ARM ID of the materialization store target.
    :type target: str

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_featurestore.py
            :start-after: [START configure_materialization_store]
            :end-before: [END configure_materialization_store]
            :language: Python
            :dedent: 8
            :caption: Configuring a Materialization Store
    """

    def __init__(self, type: str, target: str) -> None:  # pylint: disable=redefined-builtin
        self.type = type
        _ = AzureResourceId(target)
        self.__target = target

    @property
    def target(self) -> str:
        """Get target value

        :return: returns the ID of the target
        :rtype: str
        """
        return self.__target

    @target.setter
    def target(self, value: str) -> None:
        """Set target value

        :param value: the ID of the target
        :type value: str
        :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the value is an invalid ARM ID.
        """
        _ = AzureResourceId(value)
        self.__target = value



from functools import partial
from pathlib import Path
from typing import Callable, Union

import pytest
from devtools_testutils import AzureRecordedTestCase, is_live
from mock import mock
from pytest_mock import MockFixture
from test_utilities.utils import (
    _PYTEST_TIMEOUT_METHOD,
    assert_job_cancel,
    omit_with_wildcard,
    submit_and_cancel_new_dsl_pipeline,
)

from azure.ai.ml import Input, MLClient, dsl, load_component
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.dsl._group_decorator import group
from azure.ai.ml.entities import (
    Choice,
    Command,
    CommandComponent,
    Component,
    Environment,
    PipelineComponent,
    PipelineJob,
    Sweep,
)
from azure.ai.ml.operations._operation_orchestrator import OperationOrchestrator

from .._util import _DSL_TIMEOUT_SECOND

tests_root_dir = Path(__file__).parent.parent.parent
components_dir = tests_root_dir / "test_configs/components/"
job_input = Input(
    type=AssetTypes.URI_FILE,
    path="https://dprepdata.blob.core.windows.net/demo/Titanic.csv",
)
experiment_name = "dsl_pipeline_e2e"
common_omit_fields = [
    "properties",
    "display_name",
    "experiment_name",
    "jobs.*.componentId",
    "inputs.*.uri",
    "jobs.*._source",
    "jobs.*.properties",
    "settings._source",
    "source_job_id",
    "services",
]


def _get_component_in_first_child(_with_jobs: Union[PipelineJob, PipelineComponent], client: MLClient) -> Component:
    if not _with_jobs.jobs:
        raise ValueError("No jobs found in the pipeline")
    _result = next(iter(_with_jobs.jobs.values())).component.split(":")
    if len(_result) == 2:
        _name, _version = _result
    elif len(_result) == 3:
        _, _name, _version = _result
    else:
        raise ValueError("Invalid component arm string: {}".format(_result))
    return client.components.get(_name, _version)


@group
class SubGroup:
    integer: int = None
    number: float = None


@group
class Group:
    number: float = None
    sub1: SubGroup = None
    sub2: SubGroup = None


@pytest.mark.usefixtures(
    "enable_environment_id_arm_expansion",
    "enable_pipeline_private_preview_features",
    "mock_code_hash",
    "mock_component_hash",
    "mock_set_headers_with_user_aml_token",
    "recorded_test",
)
@pytest.mark.timeout(timeout=_DSL_TIMEOUT_SECOND, method=_PYTEST_TIMEOUT_METHOD)
@pytest.mark.e2etest
@pytest.mark.pipeline_test
class TestDSLPipelineWithSpecificNodes(AzureRecordedTestCase):
    @staticmethod
    def _generate_multi_layer_pipeline_func():
        path = "./tests/test_configs/components/helloworld_component.yml"

        @dsl.pipeline
        def pipeline_leaf(component_in_path: Input):
            component_func1 = load_component(source=path)
            component_func1(component_in_path=component_in_path, component_in_number=1)

            component_func2 = load_component(
                source=path,
                params_override=[
                    {
                        "name": "another_component_name",
                        "version": "another_component_version",
                    }
                ],
            )
            component_func2(component_in_path=component_in_path, component_in_number=1)

            component_func3 = load_component(
                source=path, params_override=[{"environment": "azureml:AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:32"}]
            )
            component_func3(component_in_path=component_in_path, component_in_number=1)

            component_func4 = load_component(source=path)
            component_func4.command += " & echo updated1"
            component_func4(component_in_path=component_in_path, component_in_number=1)

        @dsl.pipeline
        def pipeline_mid(job_in_path: Input):
            pipeline_leaf(job_in_path)
            pipeline_leaf(job_in_path)

        @dsl.pipeline
        def pipeline_root(job_in_path: Input):
            pipeline_mid(job_in_path)
            pipeline_mid(job_in_path)

        return pipeline_root

    @staticmethod
    def _generate_pipeline_func_for_concurrent_component_registration_test(shared_input):
        path = "./tests/test_configs/components/helloworld_component.yml"
        conda_file_path = "./tests/test_configs/environment/environment_files/environment.yml"

        environment = Environment(
            name="test-environment",
            conda_file=conda_file_path,
            image="mcr.microsoft.com/azureml/openmpi4.1.0-ubuntu20.04",
            version="2",  # TODO: anonymous environment has potential version conflict?
            description="This is an anonymous environment",
        )

        @dsl.pipeline
        def pipeline_leaf():
            component_func1a = load_component(source=path)
            component_func1a.environment = environment
            component_func1a.command += " & echo updated1"
            component_func1a(component_in_path=shared_input, component_in_number=1)

            component_func1b = load_component(source=path)
            component_func1b.environment = environment
            component_func1b.command += " & echo updated1"
            component_func1b(component_in_path=shared_input, component_in_number=1)

            component_func2 = load_component(source=path)
            component_func2.command += " & echo updated2"
            component_func2.environment = environment
            component_func2(component_in_path=shared_input, component_in_number=1)

            component_func3 = load_component(source=path)
            component_func3.command += " & echo updated3"
            component_func3.environment = environment
            component_func3(component_in_path=shared_input, component_in_number=1)

        # TODO: test with multiple pipelines after server-side return jobs for pipeline component
        # @dsl.pipeline
        # def pipeline_mid():
        #     pipeline_leaf()
        #     pipeline_leaf()
        #
        # @dsl.pipeline
        # def pipeline_root():
        #     pipeline_mid()
        #     pipeline_mid()

        return pipeline_leaf

    def test_dsl_pipeline_sweep_node(self, client: MLClient, randstr: Callable[[str], str]) -> None:
        yaml_file = "./tests/test_configs/components/helloworld_component.yml"

        @dsl.pipeline
        def train_with_sweep_in_pipeline(raw_data, primary_metric: str = "AUC", max_total_trials: int = 10):
            component_to_sweep: CommandComponent = load_component(source=yaml_file)
            # to check the logic to set default display name for components
            component_to_sweep.display_name = None

            cmd_node1: Command = component_to_sweep(
                component_in_number=Choice([2, 3, 4, 5]), component_in_path=raw_data
            )

            sweep_job1: Sweep = cmd_node1.sweep(
                primary_metric="AUC",  # primary_metric,
                goal="maximize",
                sampling_algorithm="random",
            )
            sweep_job1.compute = "gpu-cluster"
            sweep_job1.set_limits(max_total_trials=10)  # max_total_trials

            cmd_node2: Command = component_to_sweep(
                component_in_number=Choice([2, 3, 4, 5]), component_in_path=raw_data
            )
            sweep_job2: Sweep = cmd_node2.sweep(
                primary_metric="AUC",
                goal="minimize",
                sampling_algorithm="random",
                max_total_trials=10,
            )
            sweep_job2.compute = "gpu-cluster"

            sweep_job3: Sweep = component_to_sweep(
                component_in_number=Choice([2, 3, 4, 5]), component_in_path=raw_data
            ).sweep(
                primary_metric="accuracy",
                goal="maximize",
                sampling_algorithm="random",
                max_total_trials=10,
            )

            component_to_link = load_component(source=yaml_file, params_override=[{"name": "node_to_link"}])
            link_node = component_to_link(
                component_in_number=2, component_in_path=sweep_job1.outputs.component_out_path
            )

            return {
                "pipeline_job_best_model1": sweep_job1.outputs.component_out_path,
                "pipeline_job_best_model2": sweep_job2.outputs.component_out_path,
                "pipeline_job_best_model3": sweep_job3.outputs.component_out_path,
                "pipeline_model_test_result": link_node.outputs.component_out_path,
            }

        pipeline: PipelineJob = train_with_sweep_in_pipeline(
            raw_data=job_input, max_total_trials=100, primary_metric="accuracy"
        )
        pipeline.settings.default_compute = "cpu-cluster"
        created_pipeline = assert_job_cancel(pipeline, client)
        name, version = created_pipeline.jobs["sweep_job1"].trial.split(":")
        created_component = client.components.get(name, version)
        # keep original component display name to guarantee reuse
        assert created_component.display_name is None

    def test_dsl_pipeline_component_cache_in_resolver(self, client: MLClient) -> None:
        input_data_path = "./tests/test_configs/data/"
        pipeline_root = self._generate_multi_layer_pipeline_func()

        _submit_and_cancel = partial(
            submit_and_cancel_new_dsl_pipeline, client=client, job_in_path=Input(path=input_data_path)
        )

        def _mock_get_component_arm_id(_component: Component) -> str:
            # the logic has no diff comparing to original function other than always using show_progress=False
            # just to mock the function and check call information
            if not _component.id:
                _component._id = client.components.create_or_update(
                    _component, is_anonymous=True, show_progress=False
                ).id
            return _component.id

        with mock.patch.object(
            OperationOrchestrator, "_get_component_arm_id", side_effect=_mock_get_component_arm_id
        ) as mock_resolve:
            _submit_and_cancel(pipeline_root)
            # pipeline_leaf, pipeline_mid and 3 command components will be resolved
            assert mock_resolve.call_count == 5

        with mock.patch.object(
            OperationOrchestrator, "_get_component_arm_id", side_effect=_mock_get_component_arm_id
        ) as mock_resolve:
            _submit_and_cancel(pipeline_root)
            # no more requests to resolve components as local cache is hit
            assert mock_resolve.call_count == 0

        pipeline_job = pipeline_root(job_in_path=Input(path=input_data_path))
        pipeline_job.settings.default_compute = "cpu-cluster"
        leaf_subgraph = pipeline_job.jobs["pipeline_mid"].component.jobs["pipeline_leaf"].component
        leaf_subgraph.jobs["another_component_name"].component.command += " & echo updated2"
        with mock.patch.object(
            OperationOrchestrator, "_get_component_arm_id", side_effect=_mock_get_component_arm_id
        ) as mock_resolve:
            assert_job_cancel(pipeline_job, client)
            # updated command component and its parents (pipeline_leaf and pipeline_mid) will be resolved
            assert mock_resolve.call_count == 3

    def test_dsl_pipeline_concurrent_component_registration(self, client: MLClient, mocker: MockFixture) -> None:
        # disable on-disk cache to test concurrent component registration
        mocker.patch("azure.ai.ml._utils.utils.is_on_disk_cache_enabled", return_value=False)

        input_data_path = "./tests/test_configs/data/"
        pipeline_root = self._generate_pipeline_func_for_concurrent_component_registration_test(
            shared_input=Input(path=input_data_path)
        )

        _submit_and_cancel = partial(
            submit_and_cancel_new_dsl_pipeline,
            client=client,
        )

        treatment_pipeline_job = _submit_and_cancel(pipeline_root)

        with mock.patch("azure.ai.ml._utils.utils.is_concurrent_component_registration_enabled", return_value=False):
            base_pipeline_job = _submit_and_cancel(pipeline_root)

        # Server-side does not guarantee the same anonymous pipeline component share the same version
        # So omit name and version and do comparison layer by layer
        omit_fields = ["id", "name", "version", "creation_context", "services", "jobs.*.component"]

        base, treat = base_pipeline_job, treatment_pipeline_job
        # TODO: test with multiple pipelines after server-side return jobs for pipeline component
        for _ in range(0, 0):
            assert omit_with_wildcard(base._to_dict(), *omit_fields) == omit_with_wildcard(
                treat._to_dict(), *omit_fields
            )
            base = _get_component_in_first_child(base, client)
            treat = _get_component_in_first_child(treat, client)

        # The last layer contains the command components
        omit_fields.pop()
        assert omit_with_wildcard(base._to_dict(), *omit_fields) == omit_with_wildcard(treat._to_dict(), *omit_fields)

    @pytest.mark.parametrize(
        "group_param_input, use_remote_component",
        [
            pytest.param(
                Group(
                    number=1.0,
                    sub1=SubGroup(integer=1),
                ),
                False,
                id="strong-typed-group",
            ),
            pytest.param(
                {
                    "number": 1.0,
                    "sub1": {"integer": 1},
                },
                False,
                id="dict-group",
            ),
            pytest.param(
                Group(
                    number=1.0,
                    sub1=SubGroup(integer=1),
                ),
                True,
                id="strong-typed-group-remote",
            ),
            pytest.param(
                {
                    "number": 1.0,
                    "sub1": {"integer": 1},
                },
                True,
                id="dict-group-remote",
            ),
        ],
    )
    def test_dsl_pipeline_with_param_group_in_command_component(
        self,
        client,
        group_param_input,
        use_remote_component: bool,
        randstr: Callable[[str], str],
    ):
        command_func = load_component("./tests/test_configs/components/helloworld_component_with_parameter_group.yml")
        input_data_path = "./tests/test_configs/data/"

        if use_remote_component:
            command_func.name = randstr("component_name")
            command_func = client.components.create_or_update(
                command_func,
            )

        @dsl.pipeline
        def pipeline_with_command(input_data):
            command_node = command_func(
                component_in_path=input_data,
                component_in_group=group_param_input,
            )
            return command_node.outputs

        pipeline: PipelineJob = pipeline_with_command(
            input_data=Input(
                path=input_data_path,
                type=AssetTypes.URI_FOLDER,
            ),
        )
        pipeline.settings.default_compute = "cpu-cluster"
        created_pipeline = client.jobs.create_or_update(pipeline)
        assert created_pipeline.jobs["command_node"]._to_rest_inputs() == {
            "component_in_path": {"job_input_type": "literal", "value": "${{parent.inputs.input_data}}"},
            "component_in_group.number": {"job_input_type": "literal", "value": "1.0"},
            "component_in_group.sub1.integer": {"job_input_type": "literal", "value": "1"},
        }

    @pytest.mark.skipif(not is_live(), reason="Met some problem in recording")
    def test_dsl_pipeline_component_with_static_local_data_input(self, client: MLClient) -> None:
        path = "./tests/test_configs/components/helloworld_component.yml"
        input_data_path = "./tests/test_configs/data/"

        @dsl.pipeline()
        def pipeline_no_arg():
            component_func = load_component(source=path)
            component_func(
                component_in_path=Input(
                    path=input_data_path,
                    type=AssetTypes.URI_FOLDER,
                ),
                component_in_number=1,
            )

        @dsl.pipeline
        def pipeline_func():
            pipeline_no_arg()

        pipeline_job: PipelineJob = pipeline_func()
        pipeline_job.settings.default_compute = "cpu-cluster"
        client.jobs.create_or_update(pipeline_job)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import json
import logging
import typing

from marshmallow import Schema, ValidationError

from .core import MutableValidationResult, ValidationResultBuilder

module_logger = logging.getLogger(__name__)


class SchemaValidatableMixin:
    """The mixin class for schema validation."""

    @classmethod
    def _create_empty_validation_result(cls) -> MutableValidationResult:
        """Simply create an empty validation result

        To reduce _ValidationResultBuilder importing, which is a private class.

        :return: An empty validation result
        :rtype: MutableValidationResult
        """
        return ValidationResultBuilder.success()

    @classmethod
    def _load_with_schema(
        cls, data: typing.Any, *, context: typing.Any, raise_original_exception: bool = False, **kwargs: typing.Any
    ) -> typing.Any:
        schema = cls._create_schema_for_validation(context=context)

        try:
            return schema.load(data, **kwargs)
        except ValidationError as e:
            if raise_original_exception:
                raise e
            msg = "Trying to load data with schema failed. Data:\n%s\nError: %s" % (
                json.dumps(data, indent=4) if isinstance(data, dict) else data,
                json.dumps(e.messages, indent=4),
            )
            raise cls._create_validation_error(
                message=msg,
                no_personal_data_message=str(e),
            ) from e

    @classmethod
    # pylint: disable-next=docstring-missing-param
    def _create_schema_for_validation(cls, context: typing.Any) -> Schema:
        """Create a schema of the resource with specific context. Should be overridden by subclass.

        :return: The schema of the resource.
        :rtype: Schema.
        """
        raise NotImplementedError()

    def _default_context(self) -> dict:
        """Get the default context for schema validation. Should be overridden by subclass.

        :return: The default context for schema validation
        :rtype: dict
        """
        raise NotImplementedError()

    @property
    def _schema_for_validation(self) -> Schema:
        """Return the schema of this Resource with default context. Do not override this method.
        Override _create_schema_for_validation instead.

        :return: The schema of the resource.
        :rtype: Schema.
        """
        return self._create_schema_for_validation(context=self._default_context())

    def _dump_for_validation(self) -> typing.Dict:
        """Convert the resource to a dictionary.

        :return: Converted dictionary
        :rtype: typing.Dict
        """
        res: dict = self._schema_for_validation.dump(self)
        return res

    @classmethod
    def _create_validation_error(cls, message: str, no_personal_data_message: str) -> Exception:
        """The function to create the validation exception to raise in _try_raise and _validate when
        raise_error is True.

        Should be overridden by subclass.

        :param message: The error message containing detailed information
        :type message: str
        :param no_personal_data_message: The error message without personal data
        :type no_personal_data_message: str
        :return: The validation exception to raise
        :rtype: Exception
        """
        raise NotImplementedError()

    @classmethod
    def _try_raise(
        cls, validation_result: MutableValidationResult, *, raise_error: typing.Optional[bool] = True
    ) -> MutableValidationResult:
        return validation_result.try_raise(raise_error=raise_error, error_func=cls._create_validation_error)

    def _validate(self, raise_error: typing.Optional[bool] = False) -> MutableValidationResult:
        """Validate the resource. If raise_error is True, raise ValidationError if validation fails and log warnings if
        applicable; Else, return the validation result.

        :param raise_error: Whether to raise ValidationError if validation fails.
        :type raise_error: bool
        :return: The validation result
        :rtype: MutableValidationResult
        """
        result = self.__schema_validate()
        result.merge_with(self._customized_validate())
        return self._try_raise(result, raise_error=raise_error)

    def _customized_validate(self) -> MutableValidationResult:
        """Validate the resource with customized logic.

        Override this method to add customized validation logic.

        :return: The customized validation result
        :rtype: MutableValidationResult
        """
        return self._create_empty_validation_result()

    @classmethod
    def _get_skip_fields_in_schema_validation(
        cls,
    ) -> typing.List[str]:
        """Get the fields that should be skipped in schema validation.

        Override this method to add customized validation logic.

        :return: The fields to skip in schema validation
        :rtype: typing.List[str]
        """
        return []

    def __schema_validate(self) -> MutableValidationResult:
        """Validate the resource with the schema.

        :return: The validation result
        :rtype: MutableValidationResult
        """
        data = self._dump_for_validation()
        messages = self._schema_for_validation.validate(data)
        for skip_field in self._get_skip_fields_in_schema_validation():
            if skip_field in messages:
                del messages[skip_field]
        return ValidationResultBuilder.from_validation_messages(messages, data=data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import pytest

from azure.ai.ml import UserIdentityConfiguration
from azure.ai.ml._restclient.v2023_04_01_preview.models import UserIdentity as RestUserIdentity
from azure.ai.ml._restclient.v2024_01_01_preview.models import MLTableJobInput
from azure.ai.ml.automl import (
    ForecastingModels,
    ForecastingPrimaryMetrics,
    ShortSeriesHandlingConfiguration,
    forecasting,
)
from azure.ai.ml.constants import TabularTrainingMode
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl.tabular import ForecastingJob


@pytest.mark.automl_test
@pytest.mark.unittest
class TestAutoMLForecasting:
    def test_forecasting_task(self):
        # Create AutoML Forecasting Task
        identity = UserIdentityConfiguration()
        forecasting_job = forecasting(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="target",
            primary_metric=ForecastingPrimaryMetrics.NORMALIZED_MEAN_ABSOLUTE_ERROR,
            enable_model_explainability=True,
            validation_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/valid.csv"),
            # Job attributes
            compute="cluster-forecasting",
            name="forecasting_job",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
            identity=identity,
        )  # type: ForecastingJob
        forecasting_job.set_limits(timeout_minutes=60, max_trials=100, max_concurrent_trials=4)
        forecasting_job.set_training(
            enable_onnx_compatible_models=True, blocked_training_algorithms=[ForecastingModels.AVERAGE]
        )
        forecasting_job.set_featurization(enable_dnn_featurization=True)
        forecasting_job.set_forecast_settings(
            time_column_name="time",
            forecast_horizon=20,
            target_lags=[1, 2, 3],
            time_series_id_column_names=["tid1", "tid2", "tid2"],
            frequency="W-THU",
            target_rolling_window_size=4,
            short_series_handling_config=ShortSeriesHandlingConfiguration.DROP,
            use_stl="season",
            seasonality=3,
            features_unknown_at_forecast_time=[],
        )

        # check if stack ensemble was disabled
        assert forecasting_job.training.enable_stack_ensemble is False, "enable_stack_ensemble should default to False"

        # serialize and deserialize again and compare
        rest_obj = forecasting_job._to_rest_object()  # serialize to rest object
        assert isinstance(rest_obj.properties.identity, RestUserIdentity)
        assert isinstance(
            rest_obj.properties.task_details.training_data, MLTableJobInput
        ), "Training data is not MLTableJobInput"
        assert isinstance(
            rest_obj.properties.task_details.validation_data, MLTableJobInput
        ), "Validation data is not MLTableJobInput"

        original_obj = ForecastingJob._from_rest_object(rest_obj)  # deserialize from rest object
        assert forecasting_job == original_obj, "Conversion to/from rest object failed"
        assert (
            original_obj.primary_metric == ForecastingPrimaryMetrics.NORMALIZED_MEAN_ABSOLUTE_ERROR
        ), "Primary metric is not set"
        assert original_obj.compute == "cluster-forecasting", "Compute not set correctly"
        assert original_obj.name == "forecasting_job", "Name not set correctly"
        assert original_obj.experiment_name == "foo_exp", "Experiment name not set correctly"
        assert original_obj.tags == {"foo_tag": "bar"}, "Tags not set correctly"
        # check if the original job inputs were restored
        assert isinstance(original_obj.training_data, Input), "Training data is not Input"
        assert original_obj.training_data.type == AssetTypes.MLTABLE, "Training data type not set correctly"
        assert original_obj.identity == identity
        assert original_obj.training_data.path == "https://foo/bar/train.csv", "Training data path not set correctly"
        assert isinstance(original_obj.validation_data, Input), "Validation data is not Input"
        assert original_obj.validation_data.type == AssetTypes.MLTABLE, "Test data type not set correctly"
        assert original_obj.validation_data.path == "https://foo/bar/valid.csv", "Test data path not set correctly"
        assert original_obj.training.training_mode == None, "Training mode not set correctly"
        assert original_obj.limits.max_nodes == None, "Max nodes not set correctly"

    def test_forecasting_task_distributed_mode(self):
        # Create AutoML Forecasting Task
        forecasting_job = forecasting(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="target",
            primary_metric=ForecastingPrimaryMetrics.NORMALIZED_MEAN_ABSOLUTE_ERROR,
            enable_model_explainability=False,
            validation_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/valid.csv"),
            # Job attributes
            compute="cluster-forecasting",
            name="forecasting_job",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
        )  # type: ForecastingJob
        forecasting_job.set_limits(
            timeout_minutes=60,
            max_trials=100,
            max_concurrent_trials=4,
            max_nodes=4,
        )
        forecasting_job.set_training(
            enable_onnx_compatible_models=True,
            blocked_training_algorithms=[ForecastingModels.AVERAGE],
            training_mode=TabularTrainingMode.DISTRIBUTED,
        )
        forecasting_job.set_featurization(enable_dnn_featurization=True)
        forecasting_job.set_forecast_settings(
            time_column_name="time",
            forecast_horizon=20,
            target_lags=[1, 2, 3],
            time_series_id_column_names=["tid1", "tid2", "tid2"],
            frequency="W-THU",
            target_rolling_window_size=4,
            short_series_handling_config=ShortSeriesHandlingConfiguration.DROP,
            use_stl="season",
            seasonality=3,
            features_unknown_at_forecast_time=["a"],
        )

        # serialize and deserialize again and compare
        rest_obj = forecasting_job._to_rest_object()  # serialize to rest object
        original_obj = ForecastingJob._from_rest_object(rest_obj)  # deserialize from rest object

        assert forecasting_job == original_obj, "Conversion to/from rest object failed"
        assert original_obj.training.training_mode == TabularTrainingMode.DISTRIBUTED, "Training mode not set correctly"
        assert original_obj.limits.max_nodes == 4, "Max nodes not set correctly"

    def test_forecasting_task_non_distributed_mode(self):
        # Create AutoML Forecasting Task
        forecasting_job = forecasting(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="target",
            primary_metric=ForecastingPrimaryMetrics.NORMALIZED_MEAN_ABSOLUTE_ERROR,
            enable_model_explainability=False,
            validation_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/valid.csv"),
            # Job attributes
            compute="cluster-forecasting",
            name="forecasting_job",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
        )  # type: ForecastingJob
        forecasting_job.set_limits(
            timeout_minutes=60,
            max_trials=100,
            max_concurrent_trials=4,
        )
        forecasting_job.set_training(
            enable_onnx_compatible_models=True,
            blocked_training_algorithms=[ForecastingModels.AVERAGE],
            training_mode=TabularTrainingMode.NON_DISTRIBUTED,
        )
        forecasting_job.set_featurization(enable_dnn_featurization=True)
        forecasting_job.set_forecast_settings(
            time_column_name="time",
            forecast_horizon=20,
            target_lags=[1, 2, 3],
            time_series_id_column_names=["tid1", "tid2", "tid2"],
            frequency="W-THU",
            target_rolling_window_size=4,
            short_series_handling_config=ShortSeriesHandlingConfiguration.DROP,
            use_stl="season",
            seasonality=3,
            features_unknown_at_forecast_time=None,
        )

        # serialize and deserialize again and compare
        rest_obj = forecasting_job._to_rest_object()  # serialize to rest object
        original_obj = ForecastingJob._from_rest_object(rest_obj)  # deserialize from rest object

        assert forecasting_job == original_obj, "Conversion to/from rest object failed"
        assert (
            original_obj.training.training_mode == TabularTrainingMode.NON_DISTRIBUTED
        ), "Training mode not set correctly"
        assert original_obj.limits.max_nodes == None, "Max nodes not set correctly"

    def test_forecasting_task_auto_mode(self):
        # Create AutoML Forecasting Task
        forecasting_job = forecasting(
            training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
            target_column_name="target",
            primary_metric=ForecastingPrimaryMetrics.NORMALIZED_MEAN_ABSOLUTE_ERROR,
            enable_model_explainability=False,
            validation_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/valid.csv"),
            # Job attributes
            compute="cluster-forecasting",
            name="forecasting_job",
            experiment_name="foo_exp",
            tags={"foo_tag": "bar"},
        )  # type: ForecastingJob
        forecasting_job.set_limits(
            timeout_minutes=60,
            max_trials=100,
            max_concurrent_trials=4,
            max_nodes=4,
        )
        forecasting_job.set_training(
            enable_onnx_compatible_models=True,
            blocked_training_algorithms=[ForecastingModels.AVERAGE],
            training_mode=TabularTrainingMode.AUTO,
        )
        forecasting_job.set_featurization(enable_dnn_featurization=True)
        forecasting_job.set_forecast_settings(
            time_column_name="time",
            forecast_horizon=20,
            target_lags=[1, 2, 3],
            time_series_id_column_names=["tid1", "tid2", "tid2"],
            frequency="W-THU",
            target_rolling_window_size=4,
            short_series_handling_config=ShortSeriesHandlingConfiguration.DROP,
            use_stl="season",
            seasonality=3,
            features_unknown_at_forecast_time=["Z1"],
        )

        # serialize and deserialize again and compare
        rest_obj = forecasting_job._to_rest_object()  # serialize to rest object
        original_obj = ForecastingJob._from_rest_object(rest_obj)  # deserialize from rest object

        assert forecasting_job == original_obj, "Conversion to/from rest object failed"
        assert original_obj.training.training_mode == TabularTrainingMode.AUTO, "Training mode not set correctly"
        assert original_obj.limits.max_nodes == 4, "Max nodes not set correctly"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import copy
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, TypeVar, Union, cast, overload

from azure.ai.ml._utils.utils import is_data_binding_expression
from azure.ai.ml.constants import AssetTypes
from azure.ai.ml.constants._component import IOConstants
from azure.ai.ml.entities._assets._artifacts.data import Data
from azure.ai.ml.entities._assets._artifacts.model import Model
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._job.pipeline._pipeline_expression import PipelineExpressionMixin
from azure.ai.ml.entities._util import resolve_pipeline_parameter
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, UserErrorException, ValidationException

# avoid circular import error
if TYPE_CHECKING:
    from azure.ai.ml.entities import PipelineJob
    from azure.ai.ml.entities._builders import BaseNode

T = TypeVar("T")


def _build_data_binding(data: Union[str, "PipelineInput", "Output"]) -> Union[str, Output]:
    """Build input builders to data bindings.

    :param data: The data to build a data binding from
    :type data: Union[str, PipelineInput, Output]
    :return: A data binding string if data isn't a str, otherwise data
    :rtype: str
    """
    result: Union[str, Output] = ""

    if isinstance(data, (InputOutputBase)):
        # Build data binding when data is PipelineInput, Output
        result = data._data_binding()
    else:
        # Otherwise just return the data
        result = data
    return result


def _resolve_builders_2_data_bindings(
    data: Union[list, dict, str, "PipelineInput", "Output"]
) -> Union[dict, list, str, Output]:
    """Traverse data and build input builders inside it to data bindings.

    :param data: The bindings to resolve
    :type data: Union[list, dict, str, "PipelineInput", "Output"]
    :return:
       * A dict if data was a dict
       * A list if data was a list
       * A str otherwise
    :rtype: Union[list, dict, str]
    """
    if isinstance(data, dict):
        for key, val in data.items():
            if isinstance(val, (dict, list)):
                data[key] = _resolve_builders_2_data_bindings(val)
            else:
                data[key] = _build_data_binding(val)
        return data
    if isinstance(data, list):
        resolved_data = []
        for val in data:
            resolved_data.append(_resolve_builders_2_data_bindings(val))
        return resolved_data
    return _build_data_binding(data)


def _data_to_input(data: Union[Data, Model]) -> Input:
    """Convert a Data object to an Input object.

    :param data: The data to convert
    :type data: Data
    :return: The Input object
    :rtype: Input
    """
    if data.id:
        return Input(type=data.type, path=data.id)
    return Input(type=data.type, path=f"{data.name}:{data.version}")


class InputOutputBase(ABC):
    # TODO: refine this code, always use _data to store builder level settings and use _meta to store definition
    # TODO: when _data missing, return value from _meta

    def __init__(
        self,
        meta: Optional[Union[Input, Output]],
        data: Optional[Union[int, bool, float, str, Input, Output, "PipelineInput"]],
        default_data: Optional[Union[int, bool, float, str, Input, Output]] = None,
        **kwargs: Any,
    ):
        """Base class of input & output.

        :param meta: Metadata of this input/output, eg: type, min, max, etc.
        :type meta: Union[Input, Output]
        :param data: Actual value of input/output, None means un-configured data.
        :type data: Union[None, int, bool, float, str,
                          azure.ai.ml.Input,
                          azure.ai.ml.Output]
        :param default_data: default value of input/output, None means un-configured data.
        :type default_data: Union[None, int, bool, float, str,
                          azure.ai.ml.Input,
                          azure.ai.ml.Output]
        """
        self._meta = meta
        self._original_data = data
        self._data: Any = self._build_data(data)
        self._default_data = default_data
        self._type: str = meta.type if meta is not None else kwargs.pop("type", None)
        self._mode = self._get_mode(original_data=data, data=self._data, kwargs=kwargs)
        self._description = (
            self._data.description
            if self._data is not None and hasattr(self._data, "description") and self._data.description
            else kwargs.pop("description", None)
        )
        # TODO: remove this
        self._attribute_map: Dict = {}
        self._name: Optional[str] = ""
        self._version: Optional[str] = ""
        super(InputOutputBase, self).__init__(**kwargs)

    @abstractmethod
    def _build_data(self, data: T) -> Union[T, str, Input, "InputOutputBase"]:
        """Validate if data matches type and translate it to Input/Output acceptable type.

        :param data: The data
        :type data: T
        :return: The built data
        :rtype: Union[T, str, Input, InputOutputBase]
        """

    @abstractmethod
    def _build_default_data(self) -> None:
        """Build default data when data not configured."""

    @property
    def type(self) -> str:
        """Type of input/output.

        :return: The type
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type: Any) -> None:  # pylint: disable=redefined-builtin
        # For un-configured input/output, we build a default data entry for them.
        self._build_default_data()
        self._type = type
        if isinstance(self._data, (Input, Output)):
            self._data.type = type
        elif self._data is not None and not isinstance(
            self._data, (int, float, str)
        ):  # when type of self._data is InputOutputBase or its child class
            self._data._type = type

    @property
    def mode(self) -> Optional[str]:
        return self._mode

    @mode.setter
    def mode(self, mode: Optional[str]) -> None:
        # For un-configured input/output, we build a default data entry for them.
        self._build_default_data()
        self._mode = mode
        if isinstance(self._data, (Input, Output)):
            self._data.mode = mode
        elif self._data is not None and not isinstance(self._data, (int, float, str)):
            self._data._mode = mode

    @property
    def description(self) -> Any:
        return self._description

    @description.setter
    def description(self, description: str) -> None:
        # For un-configured input/output, we build a default data entry for them.
        self._build_default_data()
        self._description = description
        if isinstance(self._data, (Input, Output)):
            self._data.description = description
        elif self._data is not None and not isinstance(self._data, (int, float, str)):
            self._data._description = description

    @property
    def path(self) -> Optional[str]:
        # This property is introduced for static intellisense.
        if hasattr(self._data, "path"):
            if self._data is not None and not isinstance(self._data, (int, float, str)):
                res: Optional[str] = self._data.path
                return res
        msg = f"{type(self._data)} does not have path."
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.PIPELINE,
            error_category=ErrorCategory.USER_ERROR,
        )

    @path.setter
    def path(self, path: str) -> None:
        # For un-configured input/output, we build a default data entry for them.
        self._build_default_data()
        if hasattr(self._data, "path"):
            if self._data is not None and not isinstance(self._data, (int, float, str)):
                self._data.path = path
        else:
            msg = f"{type(self._data)} does not support setting path."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )

    def _data_binding(self) -> str:
        """Return data binding string representation for this input/output.

        :return: The data binding string
        :rtype: str
        """
        raise NotImplementedError()

    # Why did we have this function? It prevents the DictMixin from being applied.
    # Unclear if we explicitly do NOT want the mapping protocol to be applied to this, or it this was just
    # confirmation that it didn't at the time.
    def keys(self) -> None:
        # This property is introduced to raise catchable Exception in marshmallow mapping validation trial.
        raise TypeError(f"'{type(self).__name__}' object is not a mapping")

    def __str__(self) -> str:
        try:
            return self._data_binding()
        except AttributeError:
            return super(InputOutputBase, self).__str__()

    def __hash__(self) -> int:
        return id(self)

    @classmethod
    def _get_mode(
        cls,
        original_data: Optional[Union[int, bool, float, str, Input, Output, "PipelineInput"]],
        data: Optional[Union[int, bool, float, str, Input, Output]],
        kwargs: dict,
    ) -> Optional[str]:
        """Get mode of this input/output builder.

        :param original_data: Original value of input/output.
        :type original_data: Union[None, int, bool, float, str
                          azure.ai.ml.Input,
                          azure.ai.ml.Output,
                          azure.ai.ml.entities._job.pipeline._io.PipelineInput]
        :param data: Built input/output data.
        :type data: Union[None, int, bool, float, str
                          azure.ai.ml.Input,
                          azure.ai.ml.Output]
        :param kwargs: The kwargs
        :type kwargs: Dict
        :return: The mode
        :rtype: Optional[str]
        """
        # pipeline level inputs won't pass mode to bound node level inputs
        if isinstance(original_data, PipelineInput):
            return None
        return data.mode if data is not None and hasattr(data, "mode") else kwargs.pop("mode", None)

    @property
    def _is_primitive_type(self) -> bool:
        return self.type in IOConstants.PRIMITIVE_STR_2_TYPE


class NodeInput(InputOutputBase):
    """Define one input of a Component."""

    def __init__(
        self,
        port_name: str,
        meta: Optional[Input],
        *,
        data: Optional[Union[int, bool, float, str, Output, "PipelineInput", Input]] = None,
        # TODO: Bug Item number: 2883405
        owner: Optional[Union["BaseComponent", "PipelineJob"]] = None,  # type: ignore
        **kwargs: Any,
    ):
        """Initialize an input of a component.

        :param name: The name of the input.
        :type name: str
        :param meta: Metadata of this input, eg: type, min, max, etc.
        :type meta: Input
        :param data: The input data. Valid types include int, bool, float, str,
            Output of another component or pipeline input and Input.
            Note that the output of another component or pipeline input associated should be reachable in the scope
            of current pipeline. Input is introduced to support case like
            TODO: new examples
            component.inputs.xxx = Input(path="arm_id")
        :type data: Union[int, bool, float, str
                          azure.ai.ml.Output,
                          azure.ai.ml.Input]
        :param owner: The owner component of the input, used to calculate binding.
        :type owner: Union[azure.ai.ml.entities.BaseNode, azure.ai.ml.entities.PipelineJob]
        :param kwargs: A dictionary of additional configuration parameters.
        :type kwargs: dict
        """
        # TODO: validate data matches type in meta
        # TODO: validate supported data
        self._port_name = port_name
        self._owner = owner
        super().__init__(meta=meta, data=data, **kwargs)

    def _build_default_data(self) -> None:
        """Build default data when input not configured."""
        if self._data is None:
            self._data = Input()

    def _build_data(self, data: T) -> Union[T, str, Input, InputOutputBase]:
        """Build input data according to assigned input

        eg: node.inputs.key = data

        :param data: The data
        :type data: T
        :return: The built data
        :rtype: Union[T, str, Input, "PipelineInput", "NodeOutput"]
        """
        _data: Union[T, str, NodeOutput] = resolve_pipeline_parameter(data)
        if _data is None:
            return _data
        # Unidiomatic typecheck: Checks that data is _exactly_ this type, and not potentially a subtype
        if type(_data) is NodeInput:  # pylint: disable=unidiomatic-typecheck
            msg = "Can not bind input to another component's input."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        if isinstance(_data, (PipelineInput, NodeOutput)):
            # If value is input or output, it's a data binding, we require it have a owner so we can convert it to
            # a data binding, eg: ${{inputs.xxx}}
            if isinstance(_data, NodeOutput) and _data._owner is None:
                msg = "Setting input binding {} to output without owner is not allowed."
                raise ValidationException(
                    message=msg.format(_data),
                    no_personal_data_message=msg.format("[_data]"),
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                )
            return _data
        # for data binding case, set is_singular=False for case like "${{parent.inputs.job_in_folder}}/sample1.csv"
        if isinstance(_data, Input) or is_data_binding_expression(_data, is_singular=False):
            return _data
        if isinstance(_data, (Data, Model)):
            return _data_to_input(_data)
        # self._meta.type could be None when sub pipeline has no annotation
        if isinstance(self._meta, Input) and self._meta.type and not self._meta._is_primitive_type:
            if isinstance(_data, str):
                return Input(type=self._meta.type, path=_data)
            msg = "only path input is supported now but get {}: {}."
            raise UserErrorException(
                message=msg.format(type(_data), _data),
                no_personal_data_message=msg.format(type(_data), "[_data]"),
            )
        return _data

    def _to_job_input(self) -> Optional[Union[Input, str, Output]]:
        """convert the input to Input, this logic will change if backend contract changes."""
        result: Optional[Union[Input, str, Output]] = None

        if self._data is None:
            # None data means this input is not configured.
            result = None
        elif isinstance(self._data, (PipelineInput, NodeOutput)):
            # Build data binding when data is PipelineInput, Output
            result = Input(path=self._data._data_binding(), mode=self.mode)
        elif is_data_binding_expression(self._data):
            result = Input(path=self._data, mode=self.mode)
        else:
            data_binding = _build_data_binding(self._data)
            if is_data_binding_expression(self._data):
                result = Input(path=data_binding, mode=self.mode)
            else:
                result = data_binding
            # TODO: validate is self._data is supported

        return result

    def _data_binding(self) -> str:
        msg = "Input binding {} can only come from a pipeline, currently got {}"
        # call type(self._owner) to avoid circular import
        raise ValidationException(
            message=msg.format(self._port_name, type(self._owner)),
            target=ErrorTarget.PIPELINE,
            no_personal_data_message=msg.format("[port_name]", "[owner]"),
            error_category=ErrorCategory.USER_ERROR,
        )

    def _copy(self, owner: Any) -> "NodeInput":
        return NodeInput(
            port_name=self._port_name,
            data=self._data,
            owner=owner,
            meta=cast(Input, self._meta),
        )

    def _deepcopy(self) -> "NodeInput":
        return NodeInput(
            port_name=self._port_name,
            data=copy.copy(self._data),
            owner=self._owner,
            meta=cast(Input, self._meta),
        )

    def _get_data_owner(self) -> Optional["BaseNode"]:
        """Gets the data owner of the node

        Note: This only works for @pipeline, not for YAML pipeline.

        Note: Inner step will be returned as the owner when node's input is from sub pipeline's output.
            @pipeline
            def sub_pipeline():
                inner_node = component_func()
                return inner_node.outputs

            @pipeline
            def root_pipeline():
                pipeline_node = sub_pipeline()
                node = copy_files_component_func(input_dir=pipeline_node.outputs.output_dir)
                owner = node.inputs.input_dir._get_data_owner()
                assert owner == pipeline_node.nodes[0]

        :return: The node if Input is from another node's output. Returns None for literal value.
        :rtype: Optional[BaseNode]
        """
        from azure.ai.ml.entities import Pipeline
        from azure.ai.ml.entities._builders import BaseNode

        def _resolve_data_owner(data: Any) -> Optional["BaseNode"]:
            if isinstance(data, BaseNode) and not isinstance(data, Pipeline):
                return data
            while isinstance(data, PipelineInput):
                # for pipeline input, it's original value(can be literal value or another node's output)
                # is stored in _original_data
                return _resolve_data_owner(data._original_data)
            if isinstance(data, NodeOutput):
                if isinstance(data._owner, Pipeline):
                    # for input from subgraph's output, trace back to inner node
                    return _resolve_data_owner(data._binding_output)
                # for input from another node's output, return the node
                return _resolve_data_owner(data._owner)
            return None

        return _resolve_data_owner(self._data)


class NodeOutput(InputOutputBase, PipelineExpressionMixin):
    """Define one output of a Component."""

    def __init__(
        self,
        port_name: str,
        meta: Optional[Union[Input, Output]],
        *,
        data: Optional[Union[Output, str]] = None,
        # TODO: Bug Item number: 2883405
        owner: Optional[Union["BaseComponent", "PipelineJob"]] = None,  # type: ignore
        binding_output: Optional["NodeOutput"] = None,
        **kwargs: Any,
    ):
        """Initialize an Output of a component.

        :param port_name: The port_name of the output.
        :type port_name: str
        :param name: The name used to register NodeOutput/PipelineOutput data.
        :type name: str
        :param version: The version used to register NodeOutput/PipelineOutput data.
        :ype version: str
        :param data: The output data. Valid types include str, Output
        :type data: Union[str
                          azure.ai.ml.entities.Output]
        :param mode: The mode of the output.
        :type mode: str
        :param owner: The owner component of the output, used to calculate binding.
        :type owner: Union[azure.ai.ml.entities.BaseNode, azure.ai.ml.entities.PipelineJob]
        :param binding_output: The node output bound to pipeline output, only available for pipeline.
        :type binding_output: azure.ai.ml.entities.NodeOutput
        :param kwargs: A dictionary of additional configuration parameters.
        :type kwargs: dict
        :raises ~azure.ai.ml.exceptions.ValidationException: Raised if object cannot be successfully validated.
            Details will be provided in the error message.
        """
        # Allow inline output binding with string, eg: "component_out_path_1": "${{parents.outputs.job_out_data_1}}"
        if data is not None and not isinstance(data, (Output, str)):
            msg = "Got unexpected type for output: {}."
            raise ValidationException(
                message=msg.format(data),
                target=ErrorTarget.PIPELINE,
                no_personal_data_message=msg.format("[data]"),
            )
        super().__init__(meta=meta, data=data, **kwargs)
        self._port_name = port_name
        self._owner = owner
        self._name: Optional[str] = self._data.name if isinstance(self._data, Output) else None
        self._version: Optional[str] = self._data.version if isinstance(self._data, Output) else None

        self._assert_name_and_version()

        # store original node output to be able to trace back to inner node from a pipeline output builder.
        self._binding_output = binding_output

    @property
    def port_name(self) -> str:
        """The output port name, eg: node.outputs.port_name.

        :return: The port name
        :rtype: str
        """
        return self._port_name

    @property
    def name(self) -> Optional[str]:
        """Used in registering output data.

        :return: The output name
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Assigns the name to NodeOutput/PipelineOutput and builds data according to the name.

        :param name: The new name
        :type name: str
        """
        self._build_default_data()
        self._name = name
        if isinstance(self._data, Output):
            self._data.name = name
        elif isinstance(self._data, InputOutputBase):
            self._data._name = name
        else:
            raise UserErrorException(
                f"We support self._data of Input, Output, InputOutputBase, NodeOutput and NodeInput,"
                f"but got type: {type(self._data)}."
            )

    @property
    def version(self) -> Optional[str]:
        """Used in registering output data.

        :return: The output data
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version: str) -> None:
        """Assigns the version to NodeOutput/PipelineOutput and builds data according to the version.

        :param version: The new version
        :type version: str
        """
        self._build_default_data()
        self._version = version
        if isinstance(self._data, Output):
            self._data.version = version
        elif isinstance(self._data, InputOutputBase):
            self._data._version = version
        else:
            raise UserErrorException(
                f"We support self._data of Input, Output, InputOutputBase, NodeOutput and NodeInput,"
                f"but got type: {type(self._data)}."
            )

    @property
    def path(self) -> Any:
        # For node output path,
        if self._data is not None and hasattr(self._data, "path"):
            return self._data.path
        return None

    @path.setter
    def path(self, path: Optional[str]) -> None:
        # For un-configured output, we build a default data entry for them.
        self._build_default_data()
        if self._data is not None and hasattr(self._data, "path"):
            self._data.path = path
        else:
            # YAML job will have string output binding and do not support setting path for it.
            msg = f"{type(self._data)} does not support setting path."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )

    def _assert_name_and_version(self) -> None:
        if self.name and not (re.match("^[A-Za-z0-9_-]*$", self.name) and len(self.name) <= 255):
            raise UserErrorException(
                f"The output name {self.name} can only contain alphanumeric characters, dashes and underscores, "
                f"with a limit of 255 characters."
            )
        if self.version and not self.name:
            raise UserErrorException("Output name is required when output version is specified.")

    def _build_default_data(self) -> None:
        """Build default data when output not configured."""
        if self._data is None:
            # _meta will be None when node._component is not a Component object
            # so we just leave the type inference work to backend
            self._data = Output(type=None)  # type: ignore[call-overload]

    def _build_data(self, data: T) -> Any:
        """Build output data according to assigned input, eg: node.outputs.key = data

        :param data: The data
        :type data: T
        :return: `data`
        :rtype: T
        """
        if data is None:
            return data
        if not isinstance(data, (Output, str)):
            msg = f"{self.__class__.__name__} only allow set {Output.__name__} object, {type(data)} is not supported."
            raise ValidationException(
                message=msg,
                target=ErrorTarget.PIPELINE,
                no_personal_data_message=msg,
                error_category=ErrorCategory.USER_ERROR,
            )
        res: T = cast(T, data)
        return res

    def _to_job_output(self) -> Optional[Output]:
        """Convert the output to Output, this logic will change if backend contract changes."""
        if self._data is None:
            # None data means this output is not configured.
            result = None
        elif isinstance(self._data, str):
            result = Output(
                type=AssetTypes.URI_FOLDER, path=self._data, mode=self.mode, name=self.name, version=self.version
            )
        elif isinstance(self._data, Output):
            result = self._data
        elif isinstance(self._data, PipelineOutput):
            result = Output(
                type=AssetTypes.URI_FOLDER,
                path=self._data._data_binding(),
                mode=self.mode,
                name=self._data.name,
                version=self._data.version,
                description=self.description,
            )
        else:
            msg = "Got unexpected type for output: {}."
            raise ValidationException(
                message=msg.format(self._data),
                target=ErrorTarget.PIPELINE,
                no_personal_data_message=msg.format("[data]"),
            )
        return result

    def _data_binding(self) -> str:
        if self._owner is not None:
            return f"${{{{parent.jobs.{self._owner.name}.outputs.{self._port_name}}}}}"

        return ""

    def _copy(self, owner: Any) -> "NodeOutput":
        return NodeOutput(
            port_name=self._port_name,
            data=cast(Output, self._data),
            owner=owner,
            meta=self._meta,
        )

    def _deepcopy(self) -> "NodeOutput":
        return NodeOutput(
            port_name=self._port_name,
            data=cast(Output, copy.copy(self._data)),
            owner=self._owner,
            meta=self._meta,
            binding_output=self._binding_output,
        )


class PipelineInput(NodeInput, PipelineExpressionMixin):
    """Define one input of a Pipeline."""

    def __init__(self, name: str, meta: Optional[Input], group_names: Optional[List[str]] = None, **kwargs: Any):
        """Initialize a PipelineInput.

        :param name: The name of the input.
        :type name: str
        :param meta: Metadata of this input, eg: type, min, max, etc.
        :type meta: Input
        :param group_names: The input parameter's group names.
        :type group_names: List[str]
        """
        super(PipelineInput, self).__init__(port_name=name, meta=meta, **kwargs)
        self._group_names = group_names if group_names else []

    def result(self) -> Any:
        """Return original value of pipeline input.

        :return: The original value of pipeline input
        :rtype: Any

        Example:

        .. code-block:: python

           @pipeline
           def pipeline_func(param1):
             # node1's param1 will get actual value of param1 instead of a input binding.
             node1 = component_func(param1=param1.result())
        """

        # use this to break self loop
        original_data_cache: Set = set()
        original_data = self._original_data
        while isinstance(original_data, PipelineInput) and original_data not in original_data_cache:
            original_data_cache.add(original_data)
            original_data = original_data._original_data
        return original_data

    def __str__(self) -> str:
        return self._data_binding()

    @overload
    def _build_data(self, data: Union[Model, Data]) -> Input: ...

    @overload
    def _build_data(self, data: T) -> Any: ...

    def _build_data(self, data: Union[Model, Data, T]) -> Any:
        """Build data according to input type.

        :param data: The data
        :type data: Union[Model, Data, T]
        :return:
            * Input if data is a Model or Data
            * data otherwise
        :rtype: Union[Input, T]
        """
        if data is None:
            return data
        # Unidiomatic typecheck: Checks that data is _exactly_ this type, and not potentially a subtype
        if type(data) is NodeInput:  # pylint: disable=unidiomatic-typecheck
            msg = "Can not bind input to another component's input."
            raise ValidationException(message=msg, no_personal_data_message=msg, target=ErrorTarget.PIPELINE)
        if isinstance(data, (PipelineInput, NodeOutput)):
            # If value is input or output, it's a data binding, owner is required to convert it to
            # a data binding, eg: ${{parent.inputs.xxx}}
            if isinstance(data, NodeOutput) and data._owner is None:
                msg = "Setting input binding {} to output without owner is not allowed."
                raise ValidationException(
                    message=msg.format(data),
                    no_personal_data_message=msg.format("[data]"),
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                )
            return data
        if isinstance(data, (Data, Model)):
            # If value is Data, we convert it to an corresponding Input
            return _data_to_input(data)
        return data

    def _data_binding(self) -> str:
        full_name = "%s.%s" % (".".join(self._group_names), self._port_name) if self._group_names else self._port_name
        return f"${{{{parent.inputs.{full_name}}}}}"

    def _to_input(self) -> Optional[Union[Input, Output]]:
        """Convert pipeline input to component input for pipeline component.

        :return: The component input
        :rtype: Input
        """
        if self._data is None:
            # None data means this input is not configured.
            return self._meta
        data_type = self._data.type if isinstance(self._data, Input) else None
        # If type is asset type, return data type without default.
        # Else infer type from data and set it as default.
        if data_type and data_type.lower() in AssetTypes.__dict__.values():
            if not isinstance(self._data, (int, float, str)):
                result = Input(type=data_type, mode=self._data.mode)
        elif type(self._data) in IOConstants.PRIMITIVE_TYPE_2_STR:
            result = Input(
                type=IOConstants.PRIMITIVE_TYPE_2_STR[type(self._data)],
                default=self._data,
            )
        else:
            msg = f"Unsupported Input type {type(self._data)} detected when translate job to component."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.PIPELINE,
                error_category=ErrorCategory.USER_ERROR,
            )
        return result  # pylint: disable=possibly-used-before-assignment


class PipelineOutput(NodeOutput):
    """Define one output of a Pipeline."""

    def _to_job_output(self) -> Optional[Output]:
        result: Optional[Output] = None
        if isinstance(self._data, Output):
            # For pipeline output with type Output, always pass to backend.
            result = self._data
        elif self._data is None and self._meta and self._meta.type:
            # For un-configured pipeline output with meta, we need to return Output with accurate type,
            # so it won't default to uri_folder.
            result = Output(type=self._meta.type, mode=self._meta.mode, description=self._meta.description)
        else:
            result = super(PipelineOutput, self)._to_job_output()
        # Copy meta type to avoid built output's None type default to uri_folder.
        if self.type and result is not None and not result.type:
            result.type = self.type
        return result

    def _data_binding(self) -> str:
        return f"${{{{parent.outputs.{self._port_name}}}}}"

    def _to_output(self) -> Optional[Output]:
        """Convert pipeline output to component output for pipeline component."""
        if self._data is None:
            # None data means this input is not configured.
            return None
        if isinstance(self._meta, Output):
            return self._meta
        # Assign type directly as we didn't have primitive output type for now.
        if not isinstance(self._data, (int, float, str)):
            return Output(type=self._data.type, mode=self._data.mode)
        return Output()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from azure.ai.ml._restclient.v2021_10_01_dataplanepreview.models import (
    ResourceManagementAssetReferenceData,
    ResourceManagementAssetReferenceDetails,
)
from azure.ai.ml._schema import WorkspaceAssetReferenceSchema
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, PARAMS_OVERRIDE_KEY
from azure.ai.ml.entities._assets.asset import Asset
from azure.ai.ml.entities._util import load_from_dict


class WorkspaceAssetReference(Asset):
    """Workspace Model Reference.

    This is for SDK internal use only, might be deprecated in the future.
    :param name: Model name
    :type name: str
    :param version: Model version
    :type version: str
    :param asset_id: Model asset id
    :type version: str
    :param kwargs: A dictionary of additional configuration parameters.
    :type kwargs: dict
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        version: Optional[str] = None,
        asset_id: Optional[str] = None,
        properties: Optional[Dict] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name=name,
            version=version,
            properties=properties,
            **kwargs,
        )
        self.asset_id = asset_id

    @classmethod
    def _load(
        cls: Any,
        data: Optional[dict] = None,
        yaml_path: Optional[Union[os.PathLike, str]] = None,
        params_override: Optional[list] = None,
        **kwargs: Any,
    ) -> "WorkspaceAssetReference":
        data = data or {}
        params_override = params_override or []
        context = {
            BASE_PATH_CONTEXT_KEY: Path(yaml_path).parent if yaml_path else Path("./"),
            PARAMS_OVERRIDE_KEY: params_override,
        }
        res: WorkspaceAssetReference = load_from_dict(WorkspaceAssetReferenceSchema, data, context, **kwargs)
        return res

    def _to_rest_object(self) -> ResourceManagementAssetReferenceData:
        resource_management_details = ResourceManagementAssetReferenceDetails(
            destination_name=self.name,
            destination_version=self.version,
            source_asset_id=self.asset_id,
        )
        resource_management = ResourceManagementAssetReferenceData(properties=resource_management_details)
        return resource_management

    @classmethod
    def _from_rest_object(cls, resource_object: ResourceManagementAssetReferenceData) -> "WorkspaceAssetReference":
        resource_management = WorkspaceAssetReference(
            name=resource_object.properties.destination_name,
            version=resource_object.properties.destination_version,
            asset_id=resource_object.properties.source_asset_id,
        )

        return resource_management

    def _to_dict(self) -> Dict:
        return dict(WorkspaceAssetReferenceSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self))



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access

from typing import Any, Callable, Dict, Optional, Tuple, Union

from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml.constants._common import AssetTypes, LegacyAssetTypes
from azure.ai.ml.constants._component import ComponentSource, DataTransferBuiltinComponentUri, ExternalDataType
from azure.ai.ml.entities._builders.base_node import pipeline_node_decorator
from azure.ai.ml.entities._component.datatransfer_component import DataTransferCopyComponent
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._inputs_outputs.external_data import Database, FileSystem
from azure.ai.ml.entities._job.pipeline._component_translatable import ComponentTranslatableMixin
from azure.ai.ml.entities._job.pipeline._io import NodeOutput, PipelineInput
from azure.ai.ml.exceptions import ErrorTarget, ValidationErrorType, ValidationException

from .data_transfer import DataTransferCopy, DataTransferExport, DataTransferImport, _build_source_sink

SUPPORTED_INPUTS = [
    LegacyAssetTypes.PATH,
    AssetTypes.URI_FILE,
    AssetTypes.URI_FOLDER,
    AssetTypes.CUSTOM_MODEL,
    AssetTypes.MLFLOW_MODEL,
    AssetTypes.MLTABLE,
    AssetTypes.TRITON_MODEL,
]


def _parse_input(input_value: Union[Input, dict, str, PipelineInput, NodeOutput]) -> Tuple:
    component_input = None
    job_input: Union[Input, dict, str, PipelineInput, NodeOutput] = ""

    if isinstance(input_value, Input):
        component_input = Input(**input_value._to_dict())
        input_type = input_value.type
        if input_type in SUPPORTED_INPUTS:
            job_input = Input(**input_value._to_dict())
    elif isinstance(input_value, dict):
        # if user provided dict, we try to parse it to Input.
        # for job input, only parse for path type
        input_type = input_value.get("type", None)
        if input_type in SUPPORTED_INPUTS:
            job_input = Input(**input_value)
        component_input = Input(**input_value)
    elif isinstance(input_value, str):
        # Input bindings
        component_input = ComponentTranslatableMixin._to_input_builder_function(input_value)
        job_input = input_value
    elif isinstance(input_value, (PipelineInput, NodeOutput)):
        data: Any = None
        # datatransfer node can accept PipelineInput/NodeOutput for export task.
        if input_value._data is None or isinstance(input_value._data, Output):
            data = Input(type=input_value.type, mode=input_value.mode)
        else:
            data = input_value._data
        component_input, _ = _parse_input(data)
        job_input = input_value
    else:
        msg = (
            f"Unsupported input type: {type(input_value)}, only Input, dict, str, PipelineInput and NodeOutput are "
            f"supported."
        )
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.JOB,
            error_type=ValidationErrorType.INVALID_VALUE,
        )
    return component_input, job_input


def _parse_output(output_value: Union[Output, Dict]) -> Tuple:
    component_output = None
    job_output: Union[Output, Dict] = {}

    if isinstance(output_value, Output):
        component_output = Output(**output_value._to_dict())
        job_output = Output(**output_value._to_dict())
    elif not output_value:
        # output value can be None or empty dictionary
        # None output value will be packed into a JobOutput object with mode = ReadWriteMount & type = UriFolder
        component_output = ComponentTranslatableMixin._to_output(output_value)
        job_output = output_value
    elif isinstance(output_value, dict):  # When output value is a non-empty dictionary
        job_output = Output(**output_value)
        component_output = Output(**output_value)
    elif isinstance(output_value, str):  # When output is passed in from pipeline job yaml
        job_output = output_value
    else:
        msg = f"Unsupported output type: {type(output_value)}, only Output and dict are supported."
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.JOB,
            error_type=ValidationErrorType.INVALID_VALUE,
        )
    return component_output, job_output


def _parse_inputs_outputs(io_dict: Optional[Dict], parse_func: Callable) -> Tuple[Dict, Dict]:
    component_io_dict, job_io_dict = {}, {}
    if io_dict:
        for key, val in io_dict.items():
            component_io, job_io = parse_func(val)
            component_io_dict[key] = component_io
            job_io_dict[key] = job_io
    return component_io_dict, job_io_dict


@experimental
def copy_data(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[Dict] = None,
    display_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    compute: Optional[str] = None,
    inputs: Optional[Dict] = None,
    outputs: Optional[Dict] = None,
    is_deterministic: bool = True,
    data_copy_mode: Optional[str] = None,
    **kwargs: Any,
) -> DataTransferCopy:
    """Create a DataTransferCopy object which can be used inside dsl.pipeline as a function.

    :keyword name: The name of the job.
    :paramtype name: str
    :keyword description: Description of the job.
    :paramtype description: str
    :keyword tags: Tag dictionary. Tags can be added, removed, and updated.
    :paramtype tags: dict[str, str]
    :keyword display_name: Display name of the job.
    :paramtype display_name: str
    :keyword experiment_name:  Name of the experiment the job will be created under.
    :paramtype experiment_name: str
    :keyword compute: The compute resource the job runs on.
    :paramtype compute: str
    :keyword inputs: Mapping of inputs data bindings used in the job.
    :paramtype inputs: dict
    :keyword outputs: Mapping of outputs data bindings used in the job.
    :paramtype outputs: dict
    :keyword is_deterministic: Specify whether the command will return same output given same input.
        If a command (component) is deterministic, when use it as a node/step in a pipeline,
        it will reuse results from a previous submitted job in current workspace which has same inputs and settings.
        In this case, this step will not use any compute resource.
        Default to be True, specify is_deterministic=False if you would like to avoid such reuse behavior.
    :paramtype is_deterministic: bool
    :keyword data_copy_mode: data copy mode in copy task, possible value is "merge_with_overwrite", "fail_if_conflict".
    :paramtype data_copy_mode: str
    :return: A DataTransferCopy object.
    :rtype: ~azure.ai.ml.entities._component.datatransfer_component.DataTransferCopyComponent
    """
    inputs = inputs or {}
    outputs = outputs or {}
    component_inputs, job_inputs = _parse_inputs_outputs(inputs, parse_func=_parse_input)
    # job inputs can not be None
    job_inputs = {k: v for k, v in job_inputs.items() if v is not None}
    component_outputs, job_outputs = _parse_inputs_outputs(outputs, parse_func=_parse_output)
    component = kwargs.pop("component", None)
    if component is None:
        component = DataTransferCopyComponent(
            name=name,
            tags=tags,
            display_name=display_name,
            description=description,
            inputs=component_inputs,
            outputs=component_outputs,
            data_copy_mode=data_copy_mode,
            _source=ComponentSource.BUILDER,
            is_deterministic=is_deterministic,
            **kwargs,
        )
    data_transfer_copy_obj = DataTransferCopy(
        component=component,
        name=name,
        description=description,
        tags=tags,
        display_name=display_name,
        experiment_name=experiment_name,
        compute=compute,
        inputs=job_inputs,
        outputs=job_outputs,
        data_copy_mode=data_copy_mode,
        **kwargs,
    )
    return data_transfer_copy_obj


@experimental
@pipeline_node_decorator
def import_data(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[Dict] = None,
    display_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    compute: Optional[str] = None,
    source: Optional[Union[Dict, Database, FileSystem]] = None,
    outputs: Optional[Dict] = None,
    **kwargs: Any,
) -> DataTransferImport:
    """Create a DataTransferImport object which can be used inside dsl.pipeline.

    :keyword name: The name of the job.
    :paramtype name: str
    :keyword description: Description of the job.
    :paramtype description: str
    :keyword tags: Tag dictionary. Tags can be added, removed, and updated.
    :paramtype tags: dict[str, str]
    :keyword display_name: Display name of the job.
    :paramtype display_name: str
    :keyword experiment_name: Name of the experiment the job will be created under.
    :paramtype experiment_name: str
    :keyword compute: The compute resource the job runs on.
    :paramtype compute: str
    :keyword source: The data source of file system or database.
    :paramtype source: Union[Dict, ~azure.ai.ml.entities._inputs_outputs.external_data.Database,
        ~azure.ai.ml.entities._inputs_outputs.external_data.FileSystem]
    :keyword outputs: Mapping of outputs data bindings used in the job.
        The default will be an output port with the key "sink" and type "mltable".
    :paramtype outputs: dict
    :return: A DataTransferImport object.
    :rtype: ~azure.ai.ml.entities._job.pipeline._component_translatable.DataTransferImport
    """
    source = _build_source_sink(source)
    outputs = outputs or {"sink": Output(type=AssetTypes.MLTABLE)}
    # # job inputs can not be None
    # job_inputs = {k: v for k, v in job_inputs.items() if v is not None}
    _, job_outputs = _parse_inputs_outputs(outputs, parse_func=_parse_output)
    component = kwargs.pop("component", None)
    update_source = False
    if component is None:
        if source and source.type == ExternalDataType.DATABASE:
            component = DataTransferBuiltinComponentUri.IMPORT_DATABASE
        else:
            component = DataTransferBuiltinComponentUri.IMPORT_FILE_SYSTEM
        update_source = True

    data_transfer_import_obj = DataTransferImport(
        component=component,
        name=name,
        description=description,
        tags=tags,
        display_name=display_name,
        experiment_name=experiment_name,
        compute=compute,
        source=source,
        outputs=job_outputs,
        **kwargs,
    )
    if update_source:
        data_transfer_import_obj._source = ComponentSource.BUILTIN

    return data_transfer_import_obj


@experimental
@pipeline_node_decorator
def export_data(
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[Dict] = None,
    display_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    compute: Optional[str] = None,
    sink: Optional[Union[Dict, Database, FileSystem]] = None,
    inputs: Optional[Dict] = None,
    **kwargs: Any,
) -> DataTransferExport:
    """Create a DataTransferExport object which can be used inside dsl.pipeline.

    :keyword name: The name of the job.
    :paramtype name: str
    :keyword description: Description of the job.
    :paramtype description: str
    :keyword tags: Tag dictionary. Tags can be added, removed, and updated.
    :paramtype tags: dict[str, str]
    :keyword display_name: Display name of the job.
    :paramtype display_name: str
    :keyword experiment_name: Name of the experiment the job will be created under.
    :paramtype experiment_name: str
    :keyword compute: The compute resource the job runs on.
    :paramtype compute: str
    :keyword sink: The sink of external data and databases.
    :paramtype sink: Union[
        Dict,
        ~azure.ai.ml.entities._inputs_outputs.external_data.Database,
        ~azure.ai.ml.entities._inputs_outputs.external_data.FileSystem]
    :keyword inputs: Mapping of inputs data bindings used in the job.
    :paramtype inputs: dict
    :return: A DataTransferExport object.
    :rtype: ~azure.ai.ml.entities._job.pipeline._component_translatable.DataTransferExport
    :raises ValidationException: If sink is not provided or exporting file system is not supported.
    """
    sink = _build_source_sink(sink)
    _, job_inputs = _parse_inputs_outputs(inputs, parse_func=_parse_input)
    # job inputs can not be None
    job_inputs = {k: v for k, v in job_inputs.items() if v is not None}
    component = kwargs.pop("component", None)
    update_source = False
    if component is None:
        if sink and sink.type == ExternalDataType.DATABASE:
            component = DataTransferBuiltinComponentUri.EXPORT_DATABASE
        else:
            msg = "Sink is a required field for export data task and we don't support exporting file system for now."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.JOB,
                error_type=ValidationErrorType.INVALID_VALUE,
            )
        update_source = True

    data_transfer_export_obj = DataTransferExport(
        component=component,
        name=name,
        description=description,
        tags=tags,
        display_name=display_name,
        experiment_name=experiment_name,
        compute=compute,
        sink=sink,
        inputs=job_inputs,
        **kwargs,
    )
    if update_source:
        data_transfer_export_obj._source = ComponentSource.BUILTIN

    return data_transfer_export_obj



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import List, Optional

from azure.ai.ml._restclient.v2024_10_01_preview.models import ListWorkspaceKeysResult


class ContainerRegistryCredential:
    """Key for ACR associated with given workspace.

    :param location:  Location of the ACR
    :type location: str
    :param username: Username of the ACR
    :type username: str
    :param passwords: Passwords to access the ACR
    :type passwords: List[str]
    """

    def __init__(
        self, *, location: Optional[str] = None, username: Optional[str] = None, passwords: Optional[List[str]] = None
    ):
        self.location = location
        self.username = username
        self.passwords = passwords


class NotebookAccessKeys:
    """Key for notebook resource associated with given workspace.

    :param primary_access_key:  Primary access key of notebook resource
    :type primary_access_key: str
    :param secondary_access_key: Secondary access key of notebook resource
    :type secondary_access_key: str
    """

    def __init__(self, *, primary_access_key: Optional[str] = None, secondary_access_key: Optional[str] = None):
        self.primary_access_key = primary_access_key
        self.secondary_access_key = secondary_access_key


class WorkspaceKeys:
    """Workspace Keys.

    :param user_storage_key: Key for storage account associated with given workspace
    :type user_storage_key: str
    :param user_storage_resource_id: Resource id of storage account associated with given workspace
    :type user_storage_resource_id: str
    :param app_insights_instrumentation_key: Key for app insights associated with given workspace
    :type app_insights_instrumentation_key: str
    :param container_registry_credentials: Key for ACR associated with given workspace
    :type container_registry_credentials: ContainerRegistryCredential
    :param notebook_access_keys: Key for notebook resource associated with given workspace
    :type notebook_access_keys: NotebookAccessKeys
    """

    def __init__(
        self,
        *,
        user_storage_key: Optional[str] = None,
        user_storage_resource_id: Optional[str] = None,
        app_insights_instrumentation_key: Optional[str] = None,
        container_registry_credentials: Optional[ContainerRegistryCredential] = None,
        notebook_access_keys: Optional[NotebookAccessKeys] = None
    ):
        self.user_storage_key = user_storage_key
        self.user_storage_resource_id = user_storage_resource_id
        self.app_insights_instrumentation_key = app_insights_instrumentation_key
        self.container_registry_credentials = container_registry_credentials
        self.notebook_access_keys = notebook_access_keys

    @classmethod
    def _from_rest_object(cls, rest_obj: ListWorkspaceKeysResult) -> Optional["WorkspaceKeys"]:
        if not rest_obj:
            return None

        container_registry_credentials = None
        notebook_access_keys = None

        if hasattr(rest_obj, "container_registry_credentials") and rest_obj.container_registry_credentials is not None:
            container_registry_credentials = ContainerRegistryCredential(
                location=rest_obj.container_registry_credentials.location,
                username=rest_obj.container_registry_credentials.username,
                passwords=rest_obj.container_registry_credentials.passwords,
            )

        if hasattr(rest_obj, "notebook_access_keys") and rest_obj.notebook_access_keys is not None:
            notebook_access_keys = NotebookAccessKeys(
                primary_access_key=rest_obj.notebook_access_keys.primary_access_key,
                secondary_access_key=rest_obj.notebook_access_keys.secondary_access_key,
            )

        return WorkspaceKeys(
            user_storage_key=rest_obj.user_storage_key,
            user_storage_resource_id=rest_obj.user_storage_arm_id,
            app_insights_instrumentation_key=rest_obj.app_insights_instrumentation_key,
            container_registry_credentials=container_registry_credentials,
            notebook_access_keys=notebook_access_keys,
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema import PatchedSchemaMeta


class DelayMetadataSchema(metaclass=PatchedSchemaMeta):
    days = fields.Int(required=False)
    hours = fields.Int(required=False)
    minutes = fields.Int(required=False)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._feature_set.delay_metadata import DelayMetadata

        return DelayMetadata(**data)



from .input_basic import generate_dsl_pipeline as input_basic
from .input_literal_cross_type import generate_dsl_pipeline as input_literal_cross_type
from .input_literal_meta import generate_dsl_pipeline as input_literal_meta
from .input_path import generate_dsl_pipeline as input_path
from .input_path_concatenate import generate_dsl_pipeline as input_path_concatenate
from .input_reason_expression import generate_dsl_pipeline as input_reason_expression
from .input_string_concatenate import generate_dsl_pipeline as input_string_concatenate
from .run_settings_compute import generate_dsl_pipeline as run_settings_compute
from .run_settings_literal import generate_dsl_pipeline as run_settings_literal
from .run_settings_sweep_choice import generate_dsl_pipeline as run_settings_sweep_choice
from .run_settings_sweep_limits import generate_dsl_pipeline as run_settings_sweep_limits
from .run_settings_sweep_literal import generate_dsl_pipeline as run_settings_sweep_literal



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from abc import ABC
from typing import Any, Optional, Union, cast

from azure.ai.ml._restclient.v2023_08_01_preview.models import (
    BayesianSamplingAlgorithm as RestBayesianSamplingAlgorithm,
)
from azure.ai.ml._restclient.v2023_08_01_preview.models import GridSamplingAlgorithm as RestGridSamplingAlgorithm
from azure.ai.ml._restclient.v2023_08_01_preview.models import RandomSamplingAlgorithm as RestRandomSamplingAlgorithm
from azure.ai.ml._restclient.v2023_08_01_preview.models import SamplingAlgorithm as RestSamplingAlgorithm
from azure.ai.ml._restclient.v2023_08_01_preview.models import SamplingAlgorithmType
from azure.ai.ml.entities._mixins import RestTranslatableMixin


class SamplingAlgorithm(ABC, RestTranslatableMixin):
    """Base class for sampling algorithms.

    This class should not be instantiated directly. Instead, use one of its subclasses.
    """

    def __init__(self) -> None:
        self.type = None

    @classmethod
    def _from_rest_object(cls, obj: RestSamplingAlgorithm) -> Optional["SamplingAlgorithm"]:
        if not obj:
            return None

        sampling_algorithm: Any = None
        if obj.sampling_algorithm_type == SamplingAlgorithmType.RANDOM:
            sampling_algorithm = RandomSamplingAlgorithm._from_rest_object(obj)  # pylint: disable=protected-access

        if obj.sampling_algorithm_type == SamplingAlgorithmType.GRID:
            sampling_algorithm = GridSamplingAlgorithm._from_rest_object(obj)  # pylint: disable=protected-access

        if obj.sampling_algorithm_type == SamplingAlgorithmType.BAYESIAN:
            sampling_algorithm = BayesianSamplingAlgorithm._from_rest_object(obj)  # pylint: disable=protected-access

        return cast(Optional["SamplingAlgorithm"], sampling_algorithm)


class RandomSamplingAlgorithm(SamplingAlgorithm):
    """Random Sampling Algorithm.

    :keyword rule: The specific type of random algorithm. Accepted values are: "random" and "sobol".
    :type rule: str
    :keyword seed: The seed for random number generation.
    :paramtype seed: int
    :keyword logbase: A positive number or the number "e" in string format to be used as the base for log
        based random sampling.
    :paramtype logbase: Union[float, str]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_sweep_configurations.py
            :start-after: [START configure_sweep_job_random_sampling_algorithm]
            :end-before: [END configure_sweep_job_random_sampling_algorithm]
            :language: python
            :dedent: 8
            :caption: Assigning a random sampling algorithm for a SweepJob
    """

    def __init__(
        self,
        *,
        rule: Optional[str] = None,
        seed: Optional[int] = None,
        logbase: Optional[Union[float, str]] = None,
    ) -> None:
        super().__init__()
        self.type = SamplingAlgorithmType.RANDOM.lower()
        self.rule = rule
        self.seed = seed
        self.logbase = logbase

    def _to_rest_object(self) -> RestRandomSamplingAlgorithm:
        return RestRandomSamplingAlgorithm(
            rule=self.rule,
            seed=self.seed,
            logbase=self.logbase,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestRandomSamplingAlgorithm) -> "RandomSamplingAlgorithm":
        return cls(
            rule=obj.rule,
            seed=obj.seed,
            logbase=obj.logbase,
        )


class GridSamplingAlgorithm(SamplingAlgorithm):
    """Grid Sampling Algorithm.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_sweep_configurations.py
            :start-after: [START configure_sweep_job_grid_sampling_algorithm]
            :end-before: [END configure_sweep_job_grid_sampling_algorithm]
            :language: python
            :dedent: 8
            :caption: Assigning a grid sampling algorithm for a SweepJob
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = SamplingAlgorithmType.GRID.lower()

    def _to_rest_object(self) -> RestGridSamplingAlgorithm:
        return RestGridSamplingAlgorithm()

    @classmethod
    def _from_rest_object(cls, obj: RestGridSamplingAlgorithm) -> "GridSamplingAlgorithm":
        return cls()


class BayesianSamplingAlgorithm(SamplingAlgorithm):
    """Bayesian Sampling Algorithm.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_sweep_configurations.py
            :start-after: [START configure_sweep_job_bayesian_sampling_algorithm]
            :end-before: [END configure_sweep_job_bayesian_sampling_algorithm]
            :language: python
            :dedent: 8
            :caption: Assigning a Bayesian sampling algorithm for a SweepJob
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = SamplingAlgorithmType.BAYESIAN.lower()

    def _to_rest_object(self) -> RestBayesianSamplingAlgorithm:
        return RestBayesianSamplingAlgorithm()

    @classmethod
    def _from_rest_object(cls, obj: RestBayesianSamplingAlgorithm) -> "BayesianSamplingAlgorithm":
        return cls()



import argparse


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--whom", type=str, default="World")
    args = parser.parse_args()

    print(f"Hello {args.whom}!")



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from typing import Any, Dict, Mapping, Optional, TYPE_CHECKING, overload

from .. import _model_base
from .._model_base import rest_field

if TYPE_CHECKING:
    from .. import models as _models


class AzureOpenAIDeployment(_model_base.Model):
    """Azure OpenAI Deployment Information.

    Readonly variables are only populated by the server, and will be ignored when sending a request.

    :ivar name: The deployment name.
    :vartype name: str
    :ivar model_name: The name of the model to deploy.
    :vartype model_name: str
    :ivar model_version: The model version to deploy.
    :vartype model_version: str
    :ivar connection_name: The name of the connection to deploy to.
    :vartype connection_name: str
    :ivar target_url: The target URL of the AOAI resource for the deployment.
    :vartype target_url: str
    :ivar id: The ARM resource id of the deployment.
    :vartype id: str
    :ivar properties: Properties of the deployment.
    :vartype properties: dict[str, str]
    :ivar tags: Tags of the deployment.
    :vartype tags: dict[str, str]
    """

    name: Optional[str] = rest_field(visibility=["read"])
    """The deployment name."""
    model_name: Optional[str] = rest_field(visibility=["read"])
    """The name of the model to deploy."""
    model_version: Optional[str] = rest_field(visibility=["read"])
    """The model version to deploy."""
    connection_name: Optional[str] = rest_field(visibility=["read"])
    """The name of the connection to deploy to."""
    target_url: Optional[str] = rest_field(visibility=["read"])
    """The target URL of the AOAI resource for the deployment."""
    id: Optional[str] = rest_field(visibility=["read"])
    """The ARM resource id of the deployment."""


class MarketplacePlan(_model_base.Model):
    """Marketplace Subscription Definition.

    Readonly variables are only populated by the server, and will be ignored when sending a request.

    :ivar publisher_id: The id of the publisher.
    :vartype publisher_id: str
    :ivar offer_id: The id of the offering associated with the plan.
    :vartype offer_id: str
    :ivar plan_id: The id of the plan.
    :vartype plan_id: str
    :ivar term_id: The term id.
    :vartype term_id: str
    """

    publisher_id: Optional[str] = rest_field(visibility=["read"])
    """The id of the publisher."""
    offer_id: Optional[str] = rest_field(visibility=["read"])
    """The id of the offering associated with the plan."""
    plan_id: Optional[str] = rest_field(visibility=["read"])
    """The id of the plan."""
    term_id: Optional[str] = rest_field(visibility=["read"])
    """The term id."""


class MarketplaceSubscription(_model_base.Model):
    """Marketplace Subscription Definition.

    Readonly variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to server.

    :ivar name: The marketplace subscription name. Required.
    :vartype name: str
    :ivar model_id: Model id for which to create marketplace subscription. Required.
    :vartype model_id: str
    :ivar marketplace_plan: The plan associated with the marketplace subscription.
    :vartype marketplace_plan: ~azure.ai.ml.entities.models.MarketplacePlan
    :ivar status: Status of the marketplace subscription. Possible values are:
     "pending_fulfillment_start", "subscribed", "unsubscribed", "suspended".
    :vartype status: str
    :ivar provisioning_state: Provisioning state of the marketplace subscription. Possible values
     are: "creating", "deleting", "succeeded", "failed", "updating", and "canceled".
    :vartype provisioning_state: str
    :ivar id: ARM resource id of the marketplace subscription.
    :vartype id: str
    """

    name: str = rest_field()
    """The marketplace subscription name. Required."""
    model_id: str = rest_field()
    """Model id for which to create marketplace subscription. Required."""
    marketplace_plan: Optional["_models.MarketplacePlan"] = rest_field(visibility=["read"])
    """The plan associated with the marketplace subscription."""
    status: Optional[str] = rest_field(visibility=["read"])
    """Status of the marketplace subscription. Possible values are: \"pending_fulfillment_start\",
     \"subscribed\", \"unsubscribed\", \"suspended\"."""
    provisioning_state: Optional[str] = rest_field(visibility=["read"])
    """Provisioning state of the marketplace subscription. Possible values are: \"creating\",
     \"deleting\", \"succeeded\", \"failed\", \"updating\", and \"canceled\"."""
    id: Optional[str] = rest_field(visibility=["read"])
    """ARM resource id of the marketplace subscription."""

    @overload
    def __init__(
        self,
        *,
        name: str,
        model_id: str,
    ): ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]):
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint: disable=useless-super-delegation
        super().__init__(*args, **kwargs)


class ServerlessEndpoint(_model_base.Model):
    """Serverless Endpoint Definition.

    Readonly variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to server.

    :ivar name: The deployment name. Required.
    :vartype name: str
    :ivar auth_mode: Authentication mode of the endpoint.
    :vartype auth_mode: str
    :ivar model_id: The id of the model to deploy. Required.
    :vartype model_id: str
    :ivar location: Location in which to create endpoint.
    :vartype location: str
    :ivar provisioning_state: Provisioning state of the endpoint. Possible values are: "creating",
     "deleting", "succeeded", "failed", "updating", and "canceled".
    :vartype provisioning_state: str
    :ivar tags: Tags for the endpoint.
    :vartype tags: dict[str, str]
    :ivar properties: Properties of the endpoint.
    :vartype properties: dict[str, str]
    :ivar description: Descripton of the endpoint.
    :vartype description: str
    :ivar scoring_uri: Scoring uri of the endpoint.
    :vartype scoring_uri: str
    :ivar id: ARM resource id of the endpoint.
    :vartype id: str
    :ivar headers: Headers required to hit the endpoint.
    :vartype id: dict[str, str]
    """

    name: str = rest_field()
    """The deployment name. Required."""
    auth_mode: Optional[str] = rest_field()
    """Authentication mode of the endpoint. Possible values are: \"key\", \"aad\".
    Defaults to \"key\" if not given."""
    model_id: str = rest_field()
    """The id of the model to deploy. Required."""
    location: Optional[str] = rest_field(visibility=["read"])
    """Location in which to create endpoint."""
    provisioning_state: Optional[str] = rest_field(visibility=["read"])
    """Provisioning state of the endpoint. Possible values are: \"creating\", \"deleting\",
     \"succeeded\", \"failed\", \"updating\", and \"canceled\"."""
    tags: Optional[Dict[str, str]] = rest_field()
    """Tags for the endpoint."""
    properties: Optional[Dict[str, str]] = rest_field()
    """Properties of the endpoint."""
    description: Optional[str] = rest_field()
    """Descripton of the endpoint."""
    scoring_uri: Optional[str] = rest_field(visibility=["read"])
    """Scoring uri of the endpoint."""
    id: Optional[str] = rest_field(visibility=["read"])
    """ARM resource id of the endpoint."""
    headers: Optional[Dict[str, str]] = rest_field(visibility=["read"])
    """Headers required to hit the endpoint."""

    @overload
    def __init__(
        self,
        *,
        name: str,
        model_id: str,
        auth_mode: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        properties: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
    ): ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]):
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pylint: disable=useless-super-delegation
        super().__init__(*args, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from os import path
from typing import Any, Dict

from azure.ai.ml._utils.utils import load_json
from azure.ai.ml.constants._common import ArmConstants, AzureMLResourceType
from azure.ai.ml.exceptions import ErrorTarget, ValidationException

template_mapping = {
    ArmConstants.BASE_TYPE: "base_template.json",
    ArmConstants.CODE_TYPE: "code.json",
    ArmConstants.CODE_VERSION_TYPE: "code_version.json",
    ArmConstants.ENVIRONMENT_VERSION_TYPE: "environment_version.json",
    ArmConstants.MODEL_VERSION_TYPE: "model_version.json",
    ArmConstants.MODEL_TYPE: "model.json",
    ArmConstants.ONLINE_ENDPOINT_TYPE: "online_endpoint.json",
    ArmConstants.ONLINE_DEPLOYMENT_TYPE: "online_deployment.json",
    ArmConstants.UPDATE_ONLINE_ENDPOINT_TYPE: "update_online_endpoint.json",
    ArmConstants.WORKSPACE_BASE: "workspace_base.json",
    ArmConstants.WORKSPACE_PARAM: "workspace_param.json",
    ArmConstants.FEATURE_STORE_ROLE_ASSIGNMENTS: "feature_store_role_assignments.json",
    ArmConstants.FEATURE_STORE_ROLE_ASSIGNMENTS_PARAM: "feature_store_role_assignments_param.json",
    ArmConstants.WORKSPACE_PROJECT: "workspace_project.json",
}


deployment_message_mapping = {
    ArmConstants.CODE_TYPE: "Registering code: ({0})",
    AzureMLResourceType.CODE: "Registering code version: ({0})",
    ArmConstants.CODE_VERSION_TYPE: "Registering code version: ({0})",
    ArmConstants.ENVIRONMENT_VERSION_TYPE: "Registering environment version: ({0})",
    AzureMLResourceType.ENVIRONMENT: "Registering environment version: ({0})",
    ArmConstants.MODEL_VERSION_TYPE: "Registering model version: ({0})",
    ArmConstants.MODEL_TYPE: "Registering model: ({0})",
    AzureMLResourceType.MODEL: "Registering model version: ({0})",
    ArmConstants.ONLINE_ENDPOINT_TYPE: "Creating endpoint: {0}",
    ArmConstants.ONLINE_DEPLOYMENT_TYPE: "Creating or updating deployment: {0}",
    AzureMLResourceType.DEPLOYMENT: "Creating or updating deployment: {0}",
    ArmConstants.UPDATE_ONLINE_ENDPOINT_TYPE: "Updating traffic",
    ArmConstants.KEY_VAULT_PARAMETER_NAME: "Creating Key Vault: ({0})",
    ArmConstants.LOG_ANALYTICS: "Creating Log Analytics Workspace: ({0})",
    ArmConstants.APP_INSIGHTS_PARAMETER_NAME: "Creating Application Insights: ({0})",
    ArmConstants.CONTAINER_REGISTRY_PARAMETER_NAME: "Creating Container Registry: ({0})",
    ArmConstants.STORAGE_ACCOUNT_PARAMETER_NAME: "Creating Storage Account: ({0})",
    AzureMLResourceType.WORKSPACE: "Creating AzureML Workspace: ({0})",
    AzureMLResourceType.CONNECTIONS: "Creating connection: ({0})",
    ArmConstants.USER_ASSIGNED_IDENTITIES: "Creating User Assigned Identities: ({0})",
    ArmConstants.ROLE_ASSIGNMENTS: "Creating Role Assignment: ({0})",
    ArmConstants.MULTIPLE_ENDPOINTS_TYPE: "Creating endpoints: ({0})",
}


def get_template(resource_type: str) -> Dict[str, Any]:
    if resource_type not in template_mapping:
        msg = "can't find the template for the resource {}".format(resource_type)
        raise ValidationException(message=msg, no_personal_data_message=msg, target=ErrorTarget.ARM_RESOURCE)
    template_path = path.join(path.dirname(__file__), "arm_templates", template_mapping[resource_type])
    return load_json(file_path=template_path)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Optional

from azure.ai.ml.entities._mixins import RestTranslatableMixin


class FeaturizationSettings(RestTranslatableMixin):
    """Base Featurization settings."""

    def __init__(
        self,
        *,
        dataset_language: Optional[str] = None,
    ):
        self.dataset_language = dataset_language

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FeaturizationSettings):
            return NotImplemented

        return self.dataset_language == other.dataset_language

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class FeaturizationSettingsType:
    NLP = "nlp"
    TABULAR = "tabular"



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: ml_samples_sweep_configurations.py
DESCRIPTION:
    These samples demonstrate different ways to configure hyperparameter sweep jobs.
USAGE:
    python ml_samples_sweep_configurations.py

"""

import os


class SweepConfigurationOptions(object):
    def ml_sweep_config(self):
        from azure.identity import DefaultAzureCredential

        from azure.ai.ml import MLClient

        subscription_id = os.environ["AZURE_SUBSCRIPTION_ID"]
        resource_group = os.environ["RESOURCE_GROUP_NAME"]
        credential = DefaultAzureCredential()
        ml_client = MLClient(credential, subscription_id, resource_group, workspace_name="test-ws1")

        cpu_cluster = ml_client.compute.get("cpu-cluster")

        from azure.ai.ml.entities import Environment

        job_env = Environment(
            name="my-env",
            image="mcr.microsoft.com/azureml/openmpi4.1.0-ubuntu20.04:latest",
        )
        job_env = ml_client.environments.create_or_update(job_env)

        # [START configure_sweep_job_choice_loguniform]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        from azure.ai.ml.sweep import Choice, LogUniform

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        job_for_sweep = job(
            kernel=LogUniform(min_value=-6, max_value=-1),
            penalty=Choice([0.9, 0.18, 0.36, 0.72]),
        )

        # [END configure_sweep_job_choice_loguniform]

        # [START configure_sweep_job_uniform]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        from azure.ai.ml.sweep import Uniform

        job_for_sweep = job(
            kernel=Uniform(min_value=0.0005, max_value=0.005),
            penalty=Uniform(min_value=0.9, max_value=0.99),
        )
        # [END configure_sweep_job_uniform]

        # [START configure_sweep_job_bandit_policy]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        from azure.ai.ml.sweep import Uniform

        job_for_sweep = job(
            kernel=Uniform(min_value=0.0005, max_value=0.005),
            penalty=Uniform(min_value=0.9, max_value=0.99),
        )

        from azure.ai.ml.sweep import BanditPolicy

        sweep_job = job_for_sweep.sweep(
            sampling_algorithm="random",
            primary_metric="best_val_acc",
            goal="Maximize",
            max_total_trials=8,
            max_concurrent_trials=4,
            early_termination_policy=BanditPolicy(slack_factor=0.15, evaluation_interval=1, delay_evaluation=10),
        )
        # [END configure_sweep_job_bandit_policy]

        # [START configure_sweep_job_median_stopping_policy]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        from azure.ai.ml.sweep import MedianStoppingPolicy, Uniform

        job_for_sweep = job(
            kernel=Uniform(min_value=0.0005, max_value=0.005),
            penalty=Uniform(min_value=0.9, max_value=0.99),
        )

        sweep_job = job_for_sweep.sweep(
            sampling_algorithm="random",
            primary_metric="best_val_acc",
            goal="Maximize",
            max_total_trials=8,
            max_concurrent_trials=4,
            early_termination_policy=MedianStoppingPolicy(delay_evaluation=5, evaluation_interval=2),
        )
        # [END configure_sweep_job_median_stopping_policy]

        # [START configure_sweep_job_bayesian_sampling_algorithm]
        from azure.ai.ml.entities import CommandJob
        from azure.ai.ml.sweep import BayesianSamplingAlgorithm, Objective, SweepJob, SweepJobLimits

        command_job = CommandJob(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        sweep = SweepJob(
            sampling_algorithm=BayesianSamplingAlgorithm(),
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"file": "top_level.csv", "mode": "ro_mount"}},  # type:ignore
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
            objective=Objective(goal="maximize", primary_metric="accuracy"),
        )
        # [END configure_sweep_job_bayesian_sampling_algorithm]

        # [START configure_sweep_job_grid_sampling_algorithm]
        from azure.ai.ml.entities import CommandJob
        from azure.ai.ml.sweep import GridSamplingAlgorithm, SweepJob, SweepJobLimits

        command_job = CommandJob(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        sweep = SweepJob(
            sampling_algorithm=GridSamplingAlgorithm(),
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"file": "top_level.csv", "mode": "ro_mount"}},  # type:ignore
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
        )
        # [END configure_sweep_job_grid_sampling_algorithm]

        # [START configure_sweep_job_random_sampling_algorithm]
        from azure.ai.ml.entities import CommandJob
        from azure.ai.ml.sweep import RandomSamplingAlgorithm, SweepJob, SweepJobLimits

        command_job = CommandJob(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        sweep = SweepJob(
            sampling_algorithm=RandomSamplingAlgorithm(seed=999, rule="sobol", logbase="e"),
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"file": "top_level.csv", "mode": "ro_mount"}},  # type:ignore
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
        )
        # [END configure_sweep_job_random_sampling_algorithm]

        # [START configure_sweep_job_truncation_selection_policy]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        from azure.ai.ml.sweep import QUniform, TruncationSelectionPolicy, Uniform

        job_for_sweep = job(
            kernel=Uniform(min_value=0.0005, max_value=0.005),
            penalty=QUniform(min_value=0.05, max_value=0.75, q=1),
        )

        sweep_job = job_for_sweep.sweep(
            sampling_algorithm="random",
            primary_metric="best_val_acc",
            goal="Maximize",
            max_total_trials=8,
            max_concurrent_trials=4,
            early_termination_policy=TruncationSelectionPolicy(delay_evaluation=5, evaluation_interval=2),
        )
        # [END configure_sweep_job_truncation_selection_policy]

        # [START configure_sweep_job_randint_normal]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        from azure.ai.ml.sweep import Normal, Randint

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        job_for_sweep = job(
            penalty=Randint(upper=5),
            kernel=Normal(mu=2.0, sigma=1.0),
        )

        # [END configure_sweep_job_randint_normal]

        # [START configure_sweep_job_lognormal_qlognormal]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        from azure.ai.ml.sweep import LogNormal, QLogNormal

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        job_for_sweep = job(
            kernel=LogNormal(mu=0.0, sigma=1.0),
            penalty=QLogNormal(mu=5.0, sigma=2.0),
        )

        # [END configure_sweep_job_lognormal_qlognormal]

        # [START configure_sweep_job_qloguniform_qnormal]
        from azure.ai.ml import command

        job = command(
            inputs=dict(kernel="linear", penalty=1.0),
            compute=cpu_cluster,
            environment=f"{job_env.name}:{job_env.version}",
            code="./scripts",
            command="python scripts/train.py --kernel $kernel --penalty $penalty",
            experiment_name="sklearn-iris-flowers",
        )

        from azure.ai.ml.sweep import QLogUniform, QNormal

        # we can reuse an existing Command Job as a function that we can apply inputs to for the sweep configurations
        job_for_sweep = job(
            penalty=QNormal(mu=2.0, sigma=1.0, q=1),
            kernel=QLogUniform(min_value=1.0, max_value=5.0),
        )

        # [END configure_sweep_job_qloguniform_qnormal]


if __name__ == "__main__":
    sample = SweepConfigurationOptions()
    sample.ml_sweep_config()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import fields

from azure.ai.ml._schema import StringTransformedEnum
from azure.ai.ml._schema.workspace import WorkspaceSchema
from azure.ai.ml.constants import WorkspaceKind


class HubSchema(WorkspaceSchema):
    # additional_workspace_storage_accounts This field exists in the API, but is unused, and thus not surfaced yet.
    kind = StringTransformedEnum(required=True, allowed_values=WorkspaceKind.HUB)
    default_resource_group = fields.Str(required=False)
    associated_workspaces = fields.List(fields.Str(), required=False, dump_only=True)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._operations import Operations
from ._deployments_operations import DeploymentsOperations
from ._providers_operations import ProvidersOperations
from ._resources_operations import ResourcesOperations
from ._resource_groups_operations import ResourceGroupsOperations
from ._tags_operations import TagsOperations
from ._deployment_operations_operations import DeploymentOperationsOperations

__all__ = [
    "Operations",
    "DeploymentsOperations",
    "ProvidersOperations",
    "ResourcesOperations",
    "ResourceGroupsOperations",
    "TagsOperations",
    "DeploymentOperationsOperations",
]



from functools import partial
from pathlib import Path

import pytest

from azure.ai.ml import dsl, load_component, load_job
from azure.ai.ml.entities import PipelineJob

from .._util import _DSL_TIMEOUT_SECOND

tests_root_dir = Path(__file__).parent.parent.parent
components_dir = tests_root_dir / "test_configs/components/"


@pytest.mark.usefixtures("enable_pipeline_private_preview_features")
@pytest.mark.timeout(_DSL_TIMEOUT_SECOND)
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestInitFinalizeJob:
    component_func = partial(
        load_component(str(components_dir / "echo_string_component.yml")),
        component_in_string="not important",
    )
    hello_world_func = load_component(str(components_dir / "helloworld_component.yml"))

    def test_init_finalize_job_from_yaml(self) -> None:
        pipeline_job = load_job("./tests/test_configs/pipeline_jobs/pipeline_job_init_finalize.yaml")
        assert pipeline_job._validate_init_finalize_job().passed
        assert pipeline_job.settings.on_init == "a"
        assert pipeline_job.settings.on_finalize == "c"
        pipeline_job_dict = pipeline_job._to_rest_object().as_dict()
        assert pipeline_job_dict["properties"]["settings"]["on_init"] == "a"
        assert pipeline_job_dict["properties"]["settings"]["on_finalize"] == "c"

    def test_init_finalize_job_from_sdk(self) -> None:
        from azure.ai.ml._internal.dsl import set_pipeline_settings
        from azure.ai.ml.dsl import pipeline

        def assert_pipeline_job_init_finalize_job(pipeline_job: PipelineJob):
            assert pipeline_job._validate_init_finalize_job().passed
            assert pipeline_job.settings.on_init == "init_job"
            assert pipeline_job.settings.on_finalize == "finalize_job"
            pipeline_job_dict = pipeline_job._to_rest_object().as_dict()
            assert pipeline_job_dict["properties"]["settings"]["on_init"] == "init_job"
            assert pipeline_job_dict["properties"]["settings"]["on_finalize"] == "finalize_job"

        # pipeline.settings.on_init/on_finalize
        @pipeline()
        def job_settings_func():
            init_job = self.component_func()  # noqa: F841
            work1 = self.component_func()  # noqa: F841
            work2 = self.component_func()  # noqa: F841
            finalize_job = self.component_func()  # noqa: F841

        pipeline1 = job_settings_func()
        pipeline1.settings.on_init = "init_job"
        pipeline1.settings.on_finalize = "finalize_job"
        assert_pipeline_job_init_finalize_job(pipeline1)

        # dsl.settings()
        @pipeline()
        def dsl_settings_func():
            init_job = self.component_func()
            work1 = self.component_func()  # noqa: F841
            work2 = self.component_func()  # noqa: F841
            finalize_job = self.component_func()  # noqa: F841
            # `set_pipeline_settings` can receive either `BaseNode` or str, both should work
            set_pipeline_settings(on_init=init_job, on_finalize="finalize_job")

        pipeline2 = dsl_settings_func()
        assert_pipeline_job_init_finalize_job(pipeline2)

        # @pipeline(on_init, on_finalize)
        @pipeline(
            on_init="init_job",
            on_finalize="finalize_job",
        )
        def in_decorator_func():
            init_job = self.component_func()  # noqa: F841
            work1 = self.component_func()  # noqa: F841
            work2 = self.component_func()  # noqa: F841
            finalize_job = self.component_func()  # noqa: F841

        pipeline3 = in_decorator_func()
        assert_pipeline_job_init_finalize_job(pipeline3)

    def test_invalid_init_finalize_job_from_yaml(self) -> None:
        pipeline_job = load_job("./tests/test_configs/pipeline_jobs/pipeline_job_init_finalize_invalid.yaml")
        validation_result = pipeline_job._validate_init_finalize_job()
        assert not validation_result.passed
        assert (
            validation_result.error_messages["settings.on_finalize"]
            == "On_finalize job should not have connection to other execution node."
        )

    def test_invalid_init_finalize_job_from_sdk(self) -> None:
        # invalid case: job name not exists
        @dsl.pipeline()
        def invalid_init_finalize_job_func():
            self.component_func()

        invalid_pipeline1 = invalid_init_finalize_job_func()
        invalid_pipeline1.settings.on_init = "init_job"
        invalid_pipeline1.settings.on_finalize = "finalize_job"
        validation_result1 = invalid_pipeline1._validate_init_finalize_job()
        assert not validation_result1.passed
        assert validation_result1.error_messages["settings.on_init"] == "On_init job name init_job not exists in jobs."
        assert (
            validation_result1.error_messages["settings.on_finalize"]
            == "On_finalize job name finalize_job not exists in jobs."
        )

        # invalid case: no normal node, on_init/on_finalize job is not isolated
        @dsl.pipeline()
        def init_finalize_with_invalid_connection_func(int_param: int, str_param: str):
            node1 = self.hello_world_func(component_in_number=int_param, component_in_path=str_param)
            node2 = self.hello_world_func(  # noqa: F841
                component_in_number=int_param,
                component_in_path=node1.outputs.component_out_path,
            )

        invalid_pipeline2 = init_finalize_with_invalid_connection_func(int_param=0, str_param="str")
        invalid_pipeline2.settings.on_init = "node2"
        invalid_pipeline2.settings.on_finalize = "node1"
        validation_result2 = invalid_pipeline2._validate_init_finalize_job()
        assert not validation_result2.passed
        assert validation_result2.error_messages["jobs"] == "No other job except for on_init/on_finalize job."
        assert (
            validation_result2.error_messages["settings.on_init"]
            == "On_init job should not have connection to other execution node."
        )
        assert (
            validation_result2.error_messages["settings.on_finalize"]
            == "On_finalize job should not have connection to other execution node."
        )

        # invalid case: call `set_pipeline_settings` out of `pipeline` decorator
        from azure.ai.ml._internal.dsl import set_pipeline_settings
        from azure.ai.ml.exceptions import UserErrorException

        with pytest.raises(UserErrorException) as e:
            set_pipeline_settings(on_init="init_job", on_finalize="finalize_job")
        assert str(e.value) == "Please call `set_pipeline_settings` inside a `pipeline` decorated function."

        # invalid case: set on_init for pipeline component
        @dsl.pipeline
        def subgraph_func():
            node = self.component_func()
            set_pipeline_settings(on_init=node)  # set on_init for subgraph (pipeline component)

        @dsl.pipeline
        def subgraph_with_init_func():
            subgraph_func()
            self.component_func()

        with pytest.raises(UserErrorException) as e:
            subgraph_with_init_func()
        assert str(e.value) == "On_init/on_finalize is not supported for pipeline component."

    def test_init_finalize_job_with_subgraph(self) -> None:
        from azure.ai.ml._internal.dsl import set_pipeline_settings

        # happy path
        @dsl.pipeline()
        def subgraph_func():
            node = self.component_func()
            node.compute = "cpu-cluster"

        @dsl.pipeline()
        def subgraph_init_finalize_job_func():
            init_job = subgraph_func()
            subgraph_work = subgraph_func()  # noqa: F841
            finalize_job = subgraph_func()
            set_pipeline_settings(on_init=init_job, on_finalize=finalize_job)

        valid_pipeline = subgraph_init_finalize_job_func()
        assert valid_pipeline._validate().passed
        assert valid_pipeline.settings.on_init == "init_job"
        assert valid_pipeline.settings.on_finalize == "finalize_job"

    def test_init_finalize_with_group(self) -> None:
        from test_configs.dsl_pipeline.pipeline_component_with_group.pipeline import pipeline_job

        assert pipeline_job._validate().passed
        assert pipeline_job.settings.on_init == "init_job"
        assert pipeline_job.settings.on_finalize == "finalize_job"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

__path__ = __import__("pkgutil").extend_path(__path__, __name__)


from ._artifacts.artifact import Artifact
from ._artifacts.code import Code
from ._artifacts.data import Data
from ._artifacts.index import Index
from ._artifacts.model import Model
from .environment import Environment
from ._artifacts._package.model_package import ModelPackage
from .workspace_asset_reference import WorkspaceAssetReference

__all__ = ["Artifact", "Model", "Code", "Data", "Index", "Environment", "WorkspaceAssetReference", "ModelPackage"]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint:disable=unreachable

from ._setup import enable_internal_components_in_pipeline
from .entities import (
    Ae365exepool,
    AISuperComputerConfiguration,
    AISuperComputerScalePolicy,
    AISuperComputerStorageReferenceConfiguration,
    Command,
    DataTransfer,
    Distributed,
    HDInsight,
    Hemera,
    InternalInput,
    ITPConfiguration,
    ITPInteractiveConfiguration,
    ITPPriorityConfiguration,
    ITPResourceConfiguration,
    ITPRetrySettings,
    Parallel,
    Pipeline,
    Scope,
    Starlite,
    TargetSelector,
)

# enable internal components if users has imported this module directly
enable_internal_components_in_pipeline()

__all__ = [
    "TargetSelector",
    "ITPInteractiveConfiguration",
    "ITPPriorityConfiguration",
    "ITPResourceConfiguration",
    "ITPRetrySettings",
    "ITPConfiguration",
    "AISuperComputerConfiguration",
    "AISuperComputerScalePolicy",
    "AISuperComputerStorageReferenceConfiguration",
    "Command",
    "Distributed",
    "Parallel",
    "Scope",
    "DataTransfer",
    "Ae365exepool",
    "HDInsight",
    "Starlite",
    "Pipeline",
    "Hemera",
    "InternalInput",
]  # pylint: disable=unreachable



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import logging
from os import PathLike
from pathlib import Path
from typing import IO, Any, AnyStr, Dict, Optional, Union, cast

from azure.ai.ml._restclient.v2023_06_01_preview.models import CreateMonitorAction, RecurrenceFrequency
from azure.ai.ml._restclient.v2023_06_01_preview.models import Schedule as RestSchedule
from azure.ai.ml._restclient.v2023_06_01_preview.models import ScheduleProperties
from azure.ai.ml._schema.monitoring.schedule import MonitorScheduleSchema
from azure.ai.ml._utils.utils import dump_yaml_to_file
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, PARAMS_OVERRIDE_KEY, ScheduleType
from azure.ai.ml.entities._mixins import RestTranslatableMixin
from azure.ai.ml.entities._monitoring.definition import MonitorDefinition
from azure.ai.ml.entities._schedule.schedule import Schedule
from azure.ai.ml.entities._schedule.trigger import CronTrigger, RecurrenceTrigger, TriggerBase
from azure.ai.ml.entities._system_data import SystemData
from azure.ai.ml.entities._util import load_from_dict

module_logger = logging.getLogger(__name__)


class MonitorSchedule(Schedule, RestTranslatableMixin):
    """Monitor schedule.

    :keyword name: The schedule name.
    :paramtype name: str
    :keyword trigger: The schedule trigger.
    :paramtype trigger: Union[~azure.ai.ml.entities.CronTrigger, ~azure.ai.ml.entities.RecurrenceTrigger]
    :keyword create_monitor: The schedule action monitor definition.
    :paramtype create_monitor: ~azure.ai.ml.entities.MonitorDefinition
    :keyword display_name: The display name of the schedule.
    :paramtype display_name: Optional[str]
    :keyword description: A description of the schedule.
    :paramtype description: Optional[str]
    :keyword tags: Tag dictionary. Tags can be added, removed, and updated.
    :paramtype tags: Optional[dict[str, str]]
    :keyword properties: The job property dictionary.
    :paramtype properties: Optional[dict[str, str]]
    """

    def __init__(
        self,
        *,
        name: str,
        trigger: Optional[Union[CronTrigger, RecurrenceTrigger]],
        create_monitor: MonitorDefinition,
        display_name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        properties: Optional[Dict] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name=name,
            trigger=trigger,
            display_name=display_name,
            description=description,
            tags=tags,
            properties=properties,
            **kwargs,
        )
        self.create_monitor = create_monitor
        self._type = ScheduleType.MONITOR

    @classmethod
    def _load(
        cls,
        data: Optional[Dict] = None,
        yaml_path: Optional[Union[PathLike, str]] = None,
        params_override: Optional[list] = None,
        **kwargs: Any,
    ) -> "MonitorSchedule":
        data = data or {}
        params_override = params_override or []
        context = {
            BASE_PATH_CONTEXT_KEY: Path(yaml_path).parent if yaml_path else Path("./"),
            PARAMS_OVERRIDE_KEY: params_override,
        }
        return cls(
            base_path=cast(Dict, context[BASE_PATH_CONTEXT_KEY]),
            **load_from_dict(MonitorScheduleSchema, data, context, **kwargs),
        )

    def _to_rest_object(self) -> RestSchedule:
        if self.tags is not None:
            tags = {
                **self.tags,
            }
        # default data window size is calculated based on the trigger frequency
        # by default 7 days if user provides incorrect recurrence frequency
        # or a cron expression
        default_data_window_size = "P7D"
        ref_data_window_size = "P14D"
        if isinstance(self.trigger, RecurrenceTrigger):
            frequency = self.trigger.frequency.lower()
            interval = self.trigger.interval
            if frequency == RecurrenceFrequency.MINUTE.lower() or frequency == RecurrenceFrequency.HOUR.lower():
                default_data_window_size = "P1D"
                ref_data_window_size = "P2D"
            elif frequency == RecurrenceFrequency.DAY.lower():
                default_data_window_size = f"P{interval}D"
                ref_data_window_size = f"P{interval * 2}D"
            elif frequency == RecurrenceFrequency.WEEK.lower():
                default_data_window_size = f"P{interval * 7}D"
                ref_data_window_size = f"P{(interval * 7) * 2}D"
            elif frequency == RecurrenceFrequency.MONTH.lower():
                default_data_window_size = f"P{interval * 30}D"
                ref_data_window_size = f"P{(interval * 30) * 2}D"

        return RestSchedule(
            properties=ScheduleProperties(
                description=self.description,
                properties=self.properties,
                tags=tags,  # pylint: disable=possibly-used-before-assignment
                action=CreateMonitorAction(
                    monitor_definition=self.create_monitor._to_rest_object(
                        default_data_window_size=default_data_window_size, ref_data_window_size=ref_data_window_size
                    )
                ),
                display_name=self.display_name,
                is_enabled=self._is_enabled,
                trigger=self.trigger._to_rest_object() if self.trigger is not None else None,
            )
        )

    def dump(self, dest: Union[str, PathLike, IO[AnyStr]], **kwargs: Any) -> None:
        """Dump the asset content into a file in YAML format.

        :param dest: The local path or file stream to write the YAML content to.
            If dest is a file path, a new file will be created.
            If dest is an open file, the file will be written to directly.
        :type dest: Union[PathLike, str, IO[AnyStr]]
        :raises FileExistsError: Raised if dest is a file path and the file already exists.
        :raises IOError: Raised if dest is an open file and the file is not writable.
        """
        path = kwargs.pop("path", None)
        yaml_serialized = self._to_dict()
        dump_yaml_to_file(dest, yaml_serialized, default_flow_style=False, path=path, **kwargs)

    def _to_dict(self) -> Dict:
        res: dict = MonitorScheduleSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
        return res

    @classmethod
    def _from_rest_object(cls, obj: RestSchedule) -> "MonitorSchedule":
        properties = obj.properties
        return cls(
            trigger=TriggerBase._from_rest_object(properties.trigger),
            create_monitor=MonitorDefinition._from_rest_object(
                properties.action.monitor_definition, tags=obj.properties.tags
            ),
            name=obj.name,
            id=obj.id,
            display_name=properties.display_name,
            description=properties.description,
            tags=properties.tags,
            properties=properties.properties,
            provisioning_state=properties.provisioning_state,
            is_enabled=properties.is_enabled,
            creation_context=SystemData._from_rest_object(obj.system_data) if obj.system_data else None,
        )

    def _create_default_monitor_definition(self) -> None:
        self.create_monitor._populate_default_signal_information()

    def _set_baseline_data_trailing_tags_for_signal(self, signal_name: str) -> None:
        if self.tags is not None:
            self.tags[f"{signal_name}.baselinedata.datarange.type"] = "Trailing"
            self.tags[f"{signal_name}.baselinedata.datarange.window_size"] = "P7D"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from .automl_job import AutoMLJobSchema
from .automl_vertical import AutoMLVerticalSchema
from .featurization_settings import FeaturizationSettingsSchema, TableFeaturizationSettingsSchema
from .forecasting_settings import ForecastingSettingsSchema
from .table_vertical.classification import AutoMLClassificationSchema
from .table_vertical.forecasting import AutoMLForecastingSchema
from .table_vertical.regression import AutoMLRegressionSchema
from .table_vertical.table_vertical import AutoMLTableVerticalSchema
from .table_vertical.table_vertical_limit_settings import AutoMLTableLimitsSchema
from .training_settings import TrainingSettingsSchema

__all__ = [
    "AutoMLJobSchema",
    "AutoMLVerticalSchema",
    "FeaturizationSettingsSchema",
    "TableFeaturizationSettingsSchema",
    "ForecastingSettingsSchema",
    "AutoMLClassificationSchema",
    "AutoMLForecastingSchema",
    "AutoMLRegressionSchema",
    "AutoMLTableVerticalSchema",
    "AutoMLTableLimitsSchema",
    "TrainingSettingsSchema",
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
from typing import Callable, Dict, Optional, TypeVar, cast

from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

T = TypeVar("T")
module_logger = logging.getLogger(__name__)


class OperationConfig(object):
    """This class is used to store common configurations that are shared across operation objects of an MLClient object.

    :param object: _description_
    :type object: _type_
    """

    def __init__(self, show_progress: bool, enable_telemetry: bool) -> None:
        self._show_progress = show_progress
        self._enable_telemetry = enable_telemetry

    @property
    def show_progress(self) -> bool:
        """Decide wether to display progress bars for long running operations.

        :return: show_progress
        :rtype: bool
        """
        return self._show_progress

    @property
    def enable_telemetry(self) -> bool:
        """Decide whether to enable telemetry for Jupyter Notebooks - telemetry cannot be enabled for other contexts.

        :return: enable_telemetry
        :rtype: bool
        """
        return self._enable_telemetry


class OperationScope(object):
    def __init__(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: Optional[str],
        registry_name: Optional[str] = None,
        workspace_id: Optional[str] = None,
        workspace_location: Optional[str] = None,
    ):
        self._subscription_id = subscription_id
        self._resource_group_name = resource_group_name
        self._workspace_name = workspace_name
        self._registry_name = registry_name
        self._workspace_id = workspace_id
        self._workspace_location = workspace_location

    @property
    def subscription_id(self) -> str:
        return self._subscription_id

    @property
    def resource_group_name(self) -> str:
        return self._resource_group_name

    @property
    def workspace_name(self) -> Optional[str]:
        return self._workspace_name

    @workspace_name.setter
    def workspace_name(self, value: str) -> None:
        self._workspace_name = value

    @property
    def registry_name(self) -> Optional[str]:
        return self._registry_name

    @registry_name.setter
    def registry_name(self, value: str) -> None:
        self._registry_name = value


class _ScopeDependentOperations(object):
    def __init__(self, operation_scope: OperationScope, operation_config: OperationConfig):
        self._operation_scope = operation_scope
        self._operation_config = operation_config
        self._scope_kwargs: Dict = {
            "resource_group_name": self._operation_scope.resource_group_name,
        }

    @property  # type: ignore
    def _workspace_name(self) -> str:
        return cast(str, self._operation_scope.workspace_name)

    @property  # type: ignore
    def _registry_name(self) -> str:
        return cast(str, self._operation_scope.registry_name)

    @property
    def _subscription_id(self) -> str:
        return self._operation_scope.subscription_id

    @property
    def _resource_group_name(self) -> str:
        return self._operation_scope.resource_group_name

    @property
    def _show_progress(self) -> bool:
        return self._operation_config.show_progress

    @property
    def _enable_telemetry(self) -> bool:
        return self._operation_config.enable_telemetry


class OperationsContainer(object):
    def __init__(self):
        self._all_operations = {}

    @property
    def all_operations(self) -> Dict:
        return self._all_operations

    def add(self, name: str, operation: _ScopeDependentOperations) -> None:
        self._all_operations[name] = operation

    def get_operation(self, resource_type: str, type_check: Callable[[T], bool]) -> T:
        if resource_type in self.all_operations:
            operation = self.all_operations[resource_type]
            from unittest.mock import MagicMock

            if isinstance(operation, MagicMock) or type_check(operation):
                return operation
            msg = f"{resource_type} operations are initialized with wrong type: {type(operation)}."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                error_category=ErrorCategory.USER_ERROR,
                target=ErrorTarget.JOB,
                error_type=ValidationErrorType.INVALID_VALUE,
            )
        msg = f"Operation {resource_type} is not available for this client."
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            error_category=ErrorCategory.USER_ERROR,
            target=ErrorTarget.JOB,
            error_type=ValidationErrorType.INVALID_VALUE,
        )
